# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
import json
import re

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type

re_chapter_date = re.compile(r'\d{4}.\d{2}.\d{2}')

# FoOlSlide Open Source online comic management software (NO LONGER MAINTAINED)
# https://github.com/FoolCode/FoOlSlide or https://github.com/chocolatkey/FoOlSlide2 (fork)

# Supported servers:
# Jaimini's Box [EN] (disabled)
# Kirei Cake [EN] (disabled)
# Le Cercle du Scan [FR] (disabled)
# Tutto Anime Manga [IT] (disabled)


class FoOlSlide(Server):
    base_url: str
    search_url: str
    mangas_url: str
    manga_url: str
    chapter_url: str

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    @staticmethod
    def decrypt(value):
        """
        Decrypt a string with a circular shift of 13 of [a-zA-Z] characters

        nopqrstuvwxyzabcdefghijklm => abcdefghijklmnopqrstuvwxyz
        """
        dvalue = ''

        for char in value:
            code = ord(char)
            if 65 <= code <= 77 or 97 <= code <= 109:
                dvalue += chr(code + 13)
            elif 78 <= code <= 90 or 110 <= code <= 122:
                dvalue += chr(code - 13)
            else:
                dvalue += char

        return dvalue

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        adult_alert = False
        if soup.find('div', class_='alert'):
            adult_alert = True

            r = self.session_post(self.manga_url.format(initial_data['slug']), data=dict(adult='true'))
            if r is None:
                return None

            soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[self.name, ],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.find('h1', class_='title').text.strip()
        if soup.find('div', class_='thumbnail'):
            data['cover'] = soup.find('div', class_='thumbnail').img.get('src')

        # Details
        for element in soup.find('div', class_='info').find_all('b'):
            label = element.text
            value = list(element.next_siblings)[0][2:]
            if label in ('Author', 'Artist'):
                data['authors'].append(value)
            elif label in ('Description', 'Synopsis', ):
                if adult_alert:
                    data['synopsis'] = '{0}\n\n{1}'.format(
                        'ALERT: This series contains mature contents and is meant to be viewed by an adult audience.',
                        value
                    )
                else:
                    data['synopsis'] = value

        # Chapters
        for element in reversed(soup.find('div', class_='list').find_all('div', class_='element')):
            a_element = element.find('div', class_='title').a

            title = a_element.text.strip()
            slug = a_element.get('href').replace(f'{self.base_url}/read/{initial_data["slug"]}/{self.lang}/', '')[:-1]

            date_match = re.search(re_chapter_date, element.find('div', class_='meta_r').text)
            if date_match:
                date = convert_date_string(date_match.group(), '%Y.%m.%d')
            else:
                date = None

            data['chapters'].append(dict(
                slug=slug,
                date=date,
                title=title,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_post(
            self.chapter_url.format(manga_slug, chapter_slug),
            data=dict(adult='true'),
            headers={
                'Accept-Encoding': 'gzip, deflate',
                'Accept-Language': '{0}-{1},{0};q=0.9,en-US;q=0.8,en;q=0.7'.format(self.lang, self.lang.upper()),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        # List of pages is available in JavaScript variable '_0x3320' or 'pages'
        # Walk in all scripts to find it
        pages = None
        for script_element in soup.find_all('script'):
            script = script_element.string
            if script is None:
                continue

            for line in script.split('\n'):
                line = line.strip()
                if line.startswith('var _0x3320'):
                    #
                    # Jaimini's Box
                    #
                    pages = line.split(';')[0].split(',')[1][1:-2]
                    # String is encrypted with a circular shift of 13 for [a-zA-Z] characters
                    pages = self.decrypt(pages)
                    # String is BASE64 encoded
                    pages = base64.b64decode(pages)
                    break
                if line.startswith('var pages'):
                    #
                    # Kirei Cake
                    #
                    pages = line.replace('var pages = ', '')[:-1]
                    break
            if pages is not None:
                pages = json.loads(pages)
                break

        if pages is None:
            return None

        data = dict(
            pages=[],
        )
        for page in pages:
            data['pages'].append(dict(
                slug=None,
                image=page['url'],
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('?')[0].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_mangas(self, page=1):
        r = self.session_get('{0}/{1}'.format(self.mangas_url, page))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.find('div', class_='series').find_all('div', class_='group'):
            a_element = element.find('div', class_='title').a

            results.append(dict(
                slug=a_element.get('href').split('/')[-2],
                name=a_element.get('title'),
            ))

        return results

    def get_latest_updates(self):
        """
        Returns latest updates
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        slugs = []
        for a_element in soup.select('.group > .title > a'):
            slug = a_element.get('href').split('/')[-2]
            if slug in slugs:
                continue
            results.append(dict(
                slug=slug,
                name=a_element.text.strip(),
            ))
            slugs.append(slug)

        return results

    def get_most_populars(self):
        """
        Returns list of all mangas
        """
        r = self.session_get(self.mangas_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        nav_buttons = soup.find_all('a', class_='gbutton')
        if nav_buttons:
            nb_pages = int(nav_buttons[0].get('href').split('/')[-2])
        else:
            nb_pages = 1

        results = []
        for index in range(nb_pages):
            results += self.get_mangas(page=index + 1)

        return results

    def search(self, term):
        r = self.session_post(
            self.search_url,
            data=dict(search=term),
            headers={
                'Accept-Encoding': 'gzip, deflate',
                'Accept-Language': '{0}-{1},{0};q=0.9,en-US;q=0.8,en;q=0.7'.format(self.lang, self.lang.upper()),
            }
        )

        if r.status_code == 200:
            soup = BeautifulSoup(r.text, 'lxml')

            results = []
            for element in soup.find('div', class_='list').find_all('div', class_='group'):
                a_element = element.find_all('div')[0].a

                results.append(dict(
                    slug=a_element.get('href').split('/')[-2],
                    name=a_element.get('title'),
                ))

            return results

        return None



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from copy import deepcopy
from gettext import gettext as _
import threading

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk

from komikku.explorer.common import get_server_default_search_filters
from komikku.explorer.search.latest_updates import ExplorerSearchStackPageLatestUpdates
from komikku.explorer.search.most_popular import ExplorerSearchStackPageMostPopular
from komikku.explorer.search.search import ExplorerSearchStackPageSearch
from komikku.explorer.search.search_global import ExplorerSearchStackPageSearchGlobal
from komikku.models import create_db_connection
from komikku.models import Manga
from komikku.models import Settings
from komikku.utils import log_error_traceback


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/explorer_search.ui')
class ExplorerSearchPage(Adw.NavigationPage):
    __gtype_name__ = 'ExplorerSearchPage'

    title_stack = Gtk.Template.Child('title_stack')
    title = Gtk.Template.Child('title')
    viewswitcher = Gtk.Template.Child('viewswitcher')
    server_website_button = Gtk.Template.Child('server_website_button')
    filters_button = Gtk.Template.Child('filters_button')

    progressbar = Gtk.Template.Child('progressbar')
    stack = Gtk.Template.Child('stack')
    searchentry = Gtk.Template.Child('searchentry')
    filters_menu_button = Gtk.Template.Child('filters_menu_button')
    search_stack = Gtk.Template.Child('search_stack')
    search_listbox = Gtk.Template.Child('search_listbox')
    search_no_results_status_page = Gtk.Template.Child('search_no_results_status_page')
    search_intro_status_page = Gtk.Template.Child('search_intro_status_page')
    most_popular_stack = Gtk.Template.Child('most_popular_stack')
    most_popular_listbox = Gtk.Template.Child('most_popular_listbox')
    most_popular_no_results_status_page = Gtk.Template.Child('most_popular_no_results_status_page')
    latest_updates_stack = Gtk.Template.Child('latest_updates_stack')
    latest_updates_listbox = Gtk.Template.Child('latest_updates_listbox')
    latest_updates_no_results_status_page = Gtk.Template.Child('latest_updates_no_results_status_page')
    viewswitcherbar = Gtk.Template.Child('viewswitcherbar')

    page = None
    requests = {}
    search_global_mode = False
    search_filters = None
    server = None

    def __init__(self, parent):
        Adw.NavigationPage.__init__(self)

        self.parent = parent
        self.window = parent.window

        self.connect('hidden', self.on_hidden)
        self.connect('shown', self.on_shown)

        self.page_changed_handler_id = self.stack.connect('notify::visible-child-name', self.on_page_changed)

        self.server_website_button.connect('clicked', self.on_server_website_button_clicked)
        self.filters_button.connect('clicked', self.on_filters_button_clicked)

        self.searchentry.connect('activate', self.search)
        self.searchentry.connect('search-changed', self.on_search_changed)
        self.searchentry.connect('stop-search', self.on_search_stopped)

        self.filters_dialog = Adw.PreferencesDialog.new()
        self.filters_dialog.set_title(_('Filters'))
        self.filters_dialog.props.presentation_mode = Adw.DialogPresentationMode.BOTTOM_SHEET

        self.window.breakpoint.add_setter(self.viewswitcherbar, 'reveal', True)
        self.window.breakpoint.add_setter(self.title_stack, 'visible-child', self.title)

        self.latest_updates_page = ExplorerSearchStackPageLatestUpdates(self)
        self.most_popular_page = ExplorerSearchStackPageMostPopular(self)
        self.search_page = ExplorerSearchStackPageSearch(self)
        self.search_global_page = ExplorerSearchStackPageSearchGlobal(self)

    def add_actions(self):
        self.search_global_page.add_actions()

    def can_page_be_updated_with_results(self, page, server_id):
        self.requests[page].remove(server_id)

        if self.window.page != self.props.tag:
            # Not in Explorer search page
            return False
        if server_id != self.server.id:
            # server_id is not the current server
            return False
        if page != self.page:
            # page is not the current page
            return False

        return True

    def clear(self):
        self.search_page.clear()
        self.most_popular_page.clear()
        self.latest_updates_page.clear()

        self.search_global_mode = False

    def init_filters_dialog(self):
        defaults = get_server_default_search_filters(self.server)
        self.search_filters = deepcopy(defaults)

        # Hide filter menu button (global search only) in searchbar
        self.filters_menu_button.set_visible(False)

        # Show or hide filters button in headerbar
        if not self.search_filters:
            self.filters_button.set_visible(False)
            return

        self.filters_button.set_visible(True)
        self.filters_button.remove_css_class('accent')

        def build_entry(group, filter_):
            def on_changed(row):
                self.search_filters[filter_['key']] = row.get_text()

                if defaults != self.search_filters:
                    self.filters_button.add_css_class('accent')
                else:
                    self.filters_button.remove_css_class('accent')

            group.set_title(filter_['name'])

            row = Adw.EntryRow(title=filter_['description'])
            row.set_use_markup(True)
            row.connect('changed', on_changed)

            group.add(row)

            return group

        def build_select_single(group, filter_):
            def on_selected(row, _param):
                position = row.get_selected()
                self.search_filters[filter_['key']] = filter_['options'][position]['key']

                if defaults != self.search_filters:
                    self.filters_button.add_css_class('accent')
                else:
                    self.filters_button.remove_css_class('accent')

            labels = Gtk.StringList()
            selected_position = 0
            for index, option in enumerate(filter_['options']):
                labels.append(option['name'])
                if option['key'] == defaults[filter_['key']]:
                    selected_position = index

            row = Adw.ComboRow(title=filter_['name'], subtitle=filter_['description'])
            row.set_use_markup(True)
            row.set_model(labels)
            row.set_selected(selected_position)
            row.connect('notify::selected', on_selected)

            group.add(row)

            return group

        def build_select_multiple(group, filter_):
            def on_active(row, _param, key):
                if row.get_active():
                    self.search_filters[filter_['key']].append(key)
                else:
                    self.search_filters[filter_['key']].remove(key)

                if defaults != self.search_filters:
                    self.filters_button.add_css_class('accent')
                else:
                    self.filters_button.remove_css_class('accent')

            group.set_title(filter_['name'])
            group.set_description(filter_['description'])

            for option in filter_['options']:
                row = Adw.SwitchRow(title=option['name'])
                row.set_use_markup(True)
                row.set_active(option['key'] in defaults[filter_['key']])
                row.connect('notify::active', on_active, option['key'])

                group.add(row)

            return group

        def build_switch(group, filter_):
            def on_active(row, _param, key):
                self.search_filters[filter_['key']] = row.get_active()

                if defaults != self.search_filters:
                    self.filters_button.add_css_class('accent')
                else:
                    self.filters_button.remove_css_class('accent')

            row = Adw.SwitchRow(title=filter_['name'])
            row.set_use_markup(True)
            row.set_subtitle(filter_['description'])
            row.set_active(defaults[filter_['key']])
            row.connect('notify::active', on_active, filter_['key'])

            group.add(row)

            return group

        # Remove a previous used page if exists
        if page := self.filters_dialog.get_visible_page():
            self.filters_dialog.remove(page)

        page = Adw.PreferencesPage(title=_('Filters'))
        self.filters_dialog.add(page)

        for index, filter_ in enumerate(self.server.filters):
            group = Adw.PreferencesGroup()

            if filter_['type'] == 'select':
                if filter_['value_type'] == 'single':
                    build_select_single(group, filter_)
                elif filter_['value_type'] == 'multiple':
                    build_select_multiple(group, filter_)
                else:
                    raise ValueError('Invalid select value_type')  # noqa: TC003
            elif filter_['type'] == 'checkbox':
                build_switch(group, filter_)
            elif filter_['type'] == 'entry':
                build_entry(group, filter_)
            else:
                raise ValueError('Invalid filter type')  # noqa: TC003

            page.add(group)

    def on_filters_button_clicked(self, _button):
        self.filters_dialog.present(self.window)

    def on_hidden(self, _page):
        # Told thread covers to stop
        self.put_covers_rendering_on_hold()

        if self.window.previous_page == self.props.tag:
            # Don't clear on navigation push
            # The page must remain unchanged if the user comes back to this page
            return

        self.clear()

    def on_manga_clicked(self, _listbox, row):
        # Told thread covers to stop
        self.put_covers_rendering_on_hold()

        if self.search_global_mode:
            self.server = getattr(row.server_data['module'], row.server_data['class_name'])()

        self.show_manga_card(row.manga_data)

    def on_page_changed(self, _stack, _param):
        # Told thread covers to stop
        self.put_covers_rendering_on_hold()

        self.page = self.stack.props.visible_child_name

        if self.page == 'most_popular':
            self.most_popular_page.populate()
        elif self.page == 'latest_updates':
            self.latest_updates_page.populate()

    def on_search_changed(self, _entry):
        if not self.searchentry.get_text().strip():
            self.search_stack.set_visible_child_name('intro')
            self.on_search_stopped(_entry)

    def on_search_stopped(self, _entry):
        if self.page == 'search' and self.search_global_mode and self.search_global_page.lock:
            self.window.add_notification(_('Search was cancelled'), timeout=2)
            self.search_global_page.cancel()

    def on_server_website_button_clicked(self, _button):
        if self.server.base_url:
            Gtk.UriLauncher.new(uri=self.server.base_url).launch()
        else:
            self.window.add_notification(_('Oops, server website URL is unknown.'), timeout=2)

    def on_shown(self, _page):
        def do_render_covers():
            if self.window.last_navigation_action != 'pop':
                return

            # Last page has been popped from the navigation stack (user comes back)
            # Recall render_covers(), some covers may not yet have been processed
            if self.page == 'search':
                if self.search_global_mode:
                    self.search_global_page.render_covers()
                else:
                    self.search_page.render_covers()
            elif self.page == 'most_popular':
                self.most_popular_page.render_covers()
            elif self.page == 'latest_updates':
                self.latest_updates_page.render_covers()

        if not Gtk.Settings.get_default().get_property('gtk-enable-animations'):
            # When animations are disabled, popped/pushed events are sent after `shown` event (bug?)
            # Use idle_add to be sure that last `popped` or `pushed` event has been received
            GLib.idle_add(do_render_covers)
        else:
            do_render_covers()

    def put_covers_rendering_on_hold(self):
        if self.page == 'search':
            if self.search_global_mode:
                self.search_global_page.thread_covers_stop_flag = True
            else:
                self.search_page.thread_covers_stop_flag = True
        elif self.page == 'most_popular':
            self.most_popular_page.thread_covers_stop_flag = True
        elif self.page == 'latest_updates':
            self.latest_updates_page.thread_covers_stop_flag = True

    def register_request(self, page):
        if page not in self.requests:
            self.requests[page] = []

        self.requests[page].append(self.server.id)

    def reinstantiate_server(self, servers):
        """Used when servers modules origin change: server variable needs to be re-instantiated"""

        if self.server is None:
            return

        for server in servers:
            if server['id'] != self.server.id:
                continue

            self.server = getattr(server['module'], server['class_name'])()
            break

    def search(self, _entry=None):
        term = self.searchentry.get_text().strip()

        if self.search_global_mode:
            self.search_global_page.search(term)
            return

        self.search_page.search(term)

    def show(self, server=None):
        self.server = server
        self.search_global_mode = server is None

        if self.search_global_mode:
            self.search_global_page.init_filters_menu()
        else:
            self.init_filters_dialog()

        if not self.search_global_mode:
            # Search, Most Popular, Latest Updates
            self.title.set_title(self.server.name)

            has_search = self.server.true_search
            has_most_popular = getattr(self.server, 'get_most_populars', None) is not None
            has_latest_updates = getattr(self.server, 'get_latest_updates', None) is not None

            with self.stack.handler_block(self.page_changed_handler_id):
                self.stack.get_page(self.stack.get_child_by_name('most_popular')).set_visible(has_most_popular)
                self.stack.get_page(self.stack.get_child_by_name('latest_updates')).set_visible(has_latest_updates)
                self.stack.get_page(self.stack.get_child_by_name('search')).set_visible(has_search)

            if has_search:
                self.searchentry.props.placeholder_text = _('Search {}').format(self.server.name)
                self.searchentry.set_text('')
                self.search_intro_status_page.set_title(_('Search for Reading'))
                if self.server.id == 'local':
                    description = _('Empty search is allowed.')
                else:
                    description = _("""Alternatively, you can look up specific comics using the syntax:

<b>id:ID from comic URL</b>""")
                self.search_intro_status_page.set_description(description)
                self.search_stack.set_visible_child_name('intro')

            if has_search:
                start_page = 'search'
            elif has_most_popular:
                start_page = 'most_popular'
            elif has_latest_updates:
                start_page = 'latest_updates'

            viewswitcher_enabled = has_search + has_most_popular + has_latest_updates > 1
            if viewswitcher_enabled:
                self.viewswitcher.set_visible(True)
                self.viewswitcherbar.set_visible(True)
                if self.viewswitcherbar.get_reveal():
                    self.title_stack.set_visible_child(self.title)
                else:
                    self.title_stack.set_visible_child(self.viewswitcher)
            else:
                self.title_stack.set_visible_child(self.title)
                self.viewswitcher.set_visible(False)
                self.viewswitcherbar.set_visible(False)
            self.server_website_button.set_visible(self.server.id != 'local')
        else:
            # Global Search (use `search` page)
            self.title.set_title(_('Global Search'))

            self.searchentry.props.placeholder_text = _('Search globally by name')
            self.searchentry.set_text('')
            self.search_intro_status_page.set_title(_('Search for Comics'))
            self.search_intro_status_page.set_description('')
            self.search_stack.set_visible_child_name('intro')
            start_page = 'search'

            self.viewswitcher.set_visible(False)
            self.viewswitcherbar.set_visible(False)
            self.server_website_button.set_visible(False)

        self.page = start_page
        self.progressbar.set_fraction(0)
        # To be sure to be notify on next page change
        self.stack.set_visible_child_name('search')
        GLib.idle_add(self.stack.set_visible_child_name, start_page)

        self.window.navigationview.push(self)

    def show_manga_card(self, manga_data, server=None):
        def run_get(server, initial_data):
            try:
                manga_data = self.server.get_manga_data(initial_data)

                if manga_data is not None:
                    GLib.idle_add(complete_get, manga_data, server)
                else:
                    GLib.idle_add(error, server)
            except Exception as e:
                user_error_message = log_error_traceback(e)
                GLib.idle_add(error, server, user_error_message)

        def run_update(server, manga_id):
            manga = Manga.get(manga_id, server)
            try:
                success, chapters_changes, synced = manga.update_full()
                if success:
                    GLib.idle_add(complete_update, manga, server, chapters_changes, synced)
                else:
                    GLib.idle_add(error, server)
            except Exception as e:
                user_error_message = log_error_traceback(e)
                GLib.idle_add(error, server, user_error_message)

        def complete_get(manga_data, server):
            if server != self.server:
                return False

            self.window.activity_indicator.set_visible(False)

            manga = Manga.new(manga_data, self.server, Settings.get_default().long_strip_detection)

            self.window.card.init(manga)

        def complete_update(manga, server, chapters_changes, _synced):
            if chapters_changes['recent_ids']:
                # Auto download new chapters
                if Settings.get_default().new_chapters_auto_download and not manga.is_local:
                    self.window.downloader.add(chapters_changes['recent_ids'], emit_signal=True)
                    self.window.downloader.start()

                self.window.library.refresh_on_manga_state_changed(manga)

            if server != self.server:
                return False

            self.window.activity_indicator.set_visible(False)

            self.window.card.init(manga)

        def error(server, message=None):
            if server != self.server:
                return False

            self.window.activity_indicator.set_visible(False)

            self.window.add_notification(message or _("Oops, failed to retrieve manga's information."), timeout=2)

            return False

        self.window.activity_indicator.set_visible(True)

        if server is not None:
            self.server = server

        # Check if selected manga is already in database
        db_conn = create_db_connection()
        record = db_conn.execute(
            'SELECT * FROM mangas WHERE slug = ? AND server_id = ?',
            (manga_data['slug'], self.server.id)
        ).fetchone()

        if record:
            thread = threading.Thread(target=run_update, args=(self.server, record['id'], ))
        else:
            thread = threading.Thread(target=run_get, args=(self.server, manga_data, ))

        thread.daemon = True
        thread.start()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from dataclasses import dataclass
import datetime
from functools import cache
from functools import cached_property
from functools import wraps
from gettext import gettext as _
import html
from io import BytesIO
import logging
import os
import re
import subprocess
import traceback

import gi
from jxlpy import JXLImagePlugin  # noqa: F401
import magic
from PIL import Image
import requests
from requests.adapters import HTTPAdapter
from requests.adapters import TimeoutSauce
from urllib3.util.retry import Retry

gi.require_version('Gdk', '4.0')
gi.require_version('Graphene', '1.0')
gi.require_version('Gtk', '4.0')

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Graphene
from gi.repository import Gtk
from gi.repository.GdkPixbuf import Pixbuf
from gi.repository.GdkPixbuf import PixbufAnimation

from komikku.consts import REQUESTS_TIMEOUT

logger = logging.getLogger('komikku')
logging.getLogger('PIL.Image').propagate = False
logging.getLogger('PIL.PngImagePlugin').propagate = False
logging.getLogger('PIL.TiffImagePlugin').propagate = False


def check_cmdline_tool(cmd):
    try:
        p = subprocess.Popen(cmd, bufsize=0, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL)
        out, _ = p.communicate()
    except Exception:
        return False, None
    else:
        return p.returncode == 0, out.decode('utf-8').strip()


def concat_images_vertically(*args: Image):
    """
    Concat vertically a list of PIL images (same width)
    """
    h = 0
    for img in args:
        h += img.height

    dst = Image.new('RGB', (args[0].width, h))

    position = 0
    for img in args:
        dst.paste(img, (0, position))
        position += img.height

    return dst


def convert_and_resize_image(buffer, width, height, keep_aspect_ratio=True, dominant_color=True, format='JPEG'):
    """Convert and resize an image (except animated images)

    :param keep_aspect_ratio: Force to keep aspect ratio of the image
    :type keep_aspect_ratio: bool

    :param dominant_color: If scaling is not allowed, use image dominant color as background color of added borders
    :type dominant_color: bool

    :return: Converted image data
    :rtype: bytes
    """

    def get_dominant_color(img):
        # Resize image to reduce number of colors
        colors = img.resize((150, 150), resample=0).getcolors(150 * 150)
        sorted_colors = sorted(colors, key=lambda t: t[0])

        return sorted_colors[-1][1]

    def remove_alpha(img):
        if img.mode not in ('P', 'RGBA'):
            return img

        img = img.convert('RGBA')
        background = Image.new('RGBA', img.size, (255, 255, 255))

        return Image.alpha_composite(background, img)

    try:
        img = Image.open(BytesIO(buffer))
    except Exception as exc:
        logger.error('Failed to open image (Pillow)', exc_info=exc)
        return None

    if img.format in ('GIF', 'WEBP') and img.is_animated:
        return buffer

    old_width, old_height = img.size
    if keep_aspect_ratio and old_width >= old_height:
        img = remove_alpha(img)

        new_ratio = height / width

        new_img = Image.new(img.mode, (old_width, int(old_width * new_ratio)), get_dominant_color(img))
        new_img.paste(img, (0, (int(old_width * new_ratio) - old_height) // 2))
        new_img.thumbnail((width, height), Image.LANCZOS)
    else:
        new_img = img.resize((width, height), Image.LANCZOS)

    new_buffer = BytesIO()
    if format == 'JPEG':
        new_img.convert('RGB').save(new_buffer, 'JPEG', quality=90)
    else:
        # Assume format supports alpha channel (transparency)
        new_img.convert('RGBA').save(new_buffer, format)
    new_img.close()

    return new_buffer.getvalue()


def folder_size(path, exclude=None):
    if not os.path.exists(path):
        return None

    cmd = ['du', '-sh', path]
    if exclude is not None:
        cmd += [f'--exclude={exclude}']
    res = subprocess.run(cmd, stdout=subprocess.PIPE, check=False)
    size = res.stdout.split()[0].decode()

    return f'{size[:-1]} {size[-1]}iB'


def get_buffer_mime_type(buffer):
    """
    Returns the MIME type of a buffer

    :param buffer: A binary string
    :type buffer: bytes

    :return: The detected MIME type, empty string otherwise
    :rtype: str
    """
    try:
        # Using python-magic module: https://github.com/ahupp/python-magic
        mime_type = magic.from_buffer(buffer[:128], mime=True)  # noqa: TC300
    except Exception:
        mime_type = ''

    # Dirty hack: fileformat detector is broken with libmagic 547
    # Known formats that cause problems: WebP
    version = magic.version()
    if version > 546 and mime_type == 'application/octet-stream':
        logger.debug('Failed to detect image file format! libmagic %d is broken?', version)
        mime_type = 'image/webp'

    return mime_type


@cache
def get_cache_dir():
    cache_dir_path = GLib.get_user_cache_dir()

    # Check if inside flatpak sandbox
    if is_flatpak():
        return cache_dir_path

    cache_dir_path = os.path.join(cache_dir_path, 'komikku')
    if not os.path.exists(cache_dir_path):
        os.mkdir(cache_dir_path)

    return cache_dir_path


@cache
def get_cached_data_dir():
    dir_path = os.path.join(get_cache_dir(), 'tmp')
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)

    return dir_path


@cache
def get_cached_logos_dir():
    dir_path = os.path.join(get_cache_dir(), 'logos')
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)

    return dir_path


@cache
def get_data_dir():
    data_dir_path = GLib.get_user_data_dir()
    app_profile = Gio.Application.get_default().profile

    if not is_flatpak():
        base_path = data_dir_path
        data_dir_path = os.path.join(base_path, 'komikku')
        if app_profile == 'development':
            data_dir_path += '-devel'
        elif app_profile == 'beta':
            data_dir_path += '-beta'

        if not os.path.exists(data_dir_path):
            os.mkdir(data_dir_path)

    # Create folder for 'local' server
    data_local_dir_path = os.path.join(data_dir_path, 'local')
    if not os.path.exists(data_local_dir_path):
        os.mkdir(data_local_dir_path)

    return data_dir_path


def get_file_mime_type(path):
    """
    Returns the MIME type of a file

    :param path: A file path
    :type path: str

    :return: The detected MIME type, empty string otherwise
    :rtype: str
    """
    try:
        # Using python-magic module: https://github.com/ahupp/python-magic
        return magic.from_file(path, mime=True)  # noqa: TC300
    except Exception:
        return ''


def get_image_info(path_or_bytes):
    try:
        if isinstance(path_or_bytes, str):
            img = Image.open(path_or_bytes)
        else:
            img = Image.open(BytesIO(path_or_bytes))
    except Exception:
        # Pillow doesn´t support SVG images
        # Get content type to identify an image
        if isinstance(path_or_bytes, str):
            gfile = Gio.File.new_for_path(path_or_bytes)
            content_type = gfile.query_info('standard::content-type', Gio.FileQueryInfoFlags.NONE, None).get_content_type()
        else:
            content_type, _result_uncertain = Gio.content_type_guess(None, path_or_bytes)

        if content_type in ('image/svg+xml',):
            info = {
                'width': -1,
                'height': -1,
                'is_animated': False,
            }
        else:
            logger.warning('Pillow: Failed to open file - corrupted or unsupported image format (%s)', content_type)
            info = None
    else:
        content_type = img.get_format_mimetype()
        info = {
            'width': img.width,
            'height': img.height,
            'is_animated': hasattr(img, 'is_animated') and img.is_animated,
        }

        img.close()

    return info, content_type


def get_response_elapsed(r):
    """
    Returns the response time (in seconds) of a request
    regardless of the request type (requests, curl_cffi)

    :param r: A response
    :type r: requests.models.Response or curl_cffi.requests.models.Response

    :return: How many seconds the request cost
    :rtype: float
    """
    elapsed = r.elapsed
    if isinstance(elapsed, datetime.timedelta):
        # requests HTTP client
        return elapsed.total_seconds()

    # curl_cffi HTTP client
    return elapsed


@cache
def get_webview_data_dir():
    return os.path.join(get_cache_dir(), 'webview')


def html_escape(s):
    return html.escape(html.unescape(s), quote=False)


def if_network_available(func_=None, only_notify=False):
    """Decorator to disable an action when network is not avaibable"""

    def _decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            window = args[0].parent if hasattr(args[0], 'parent') else args[0].window
            if not window.network_available:
                window.add_notification(_('You are currently offline'), timeout=3, priority=1)
                if not only_notify:
                    return None

            return func(*args, **kwargs)
        return wrapper

    if callable(func_):
        return _decorator(func_)
    elif func_ is None:
        return _decorator
    else:
        raise RuntimeWarning('Positional arguments are not supported')


def is_flatpak():
    return os.path.exists(os.path.join(GLib.get_user_runtime_dir(), 'flatpak-info'))


def is_number(s):
    return s is not None and str(s).replace('.', '', 1).isdigit()


def log_error_traceback(e):
    from komikku.servers.exceptions import ServerException

    if isinstance(e, requests.exceptions.RequestException):
        return _('No Internet connection, timeout or server down')
    if isinstance(e, ServerException):
        return e.message

    logger.info(traceback.format_exc())

    return None


def markdown_to_markup(s):
    # Escape HTML
    s = html_escape(s)

    # Convert links into <a> tags
    return re.sub(
        r'\[(.*?)\]\((\S*?)\s*("(.*?)")?\)',  # 1. text, 2. url, 4. title
        r'<a href="\g<2>">\g<1></a>',
        s,
        flags=re.M
    )


def remove_number_leading_zero(str_num):
    """Remove leading zero in a number string

    '00123' => '123' (int)
    '00123.45' => '123.45' (float)
    """
    return str(int(float(str_num))) if int(float(str_num)) == float(str_num) else str(float(str_num))


def retry_session(session=None, retries=3, allowed_methods=['GET'], backoff_factor=0.3, status_forcelist=None):
    if session is None:
        session = requests.Session()
    elif not getattr(session, 'adapters', None) or session.adapters['https://'].max_retries.total == retries:
        # Retry adapter is already modified or session is not a `requests (HTTP client)` session
        return session

    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        allowed_methods=allowed_methods,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)

    session.mount('http://', adapter)
    session.mount('https://', adapter)

    return session


def skip_past(haystack, needle):
    if (idx := haystack.find(needle)) >= 0:
        return idx + len(needle)

    return None


def trunc_filename(filename):
    """Reduce filename length to 255 (common FS limit) if it's too long"""
    return filename.encode('utf-8')[:255].decode().strip()


class BaseServer:
    id: str
    name: str

    description = None
    headers = None
    headers_images = None
    http_client = 'requests'  # HTTP client
    status = 'enabled'

    __sessions = {}  # to cache all existing sessions

    @property
    def session(self):
        return BaseServer.__sessions.get(self.id)

    @session.setter
    def session(self, value):
        BaseServer.__sessions[self.id] = value

    @cached_property
    def sessions_dir(self):
        dir_path = os.path.join(get_cache_dir(), 'sessions')
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)

        return dir_path

    def get_image(self, url, etag=None):
        """
        Get an image

        :param url: The image URL
        :type url: str

        :param etag: The current image ETag
        :type etag: str or None

        :return: The image content, the image ETag if exists, the request time (seconds)
        :rtype: tuple
        """
        if url is None:
            return None, None, None

        if self.headers_images is not None:
            headers = self.headers_images
        else:
            headers = {
                'Accept': 'image/avif,image/webp,*/*',
                'Referer': f'{self.base_url}/',
            }
        if etag:
            headers['If-None-Match'] = etag

        if self.session:
            r = self.session.get(url, headers=headers)
        else:
            # Session object has not yet been instantiated (servers with login or using webview to complete challenge)
            r = requests.get(url, headers=headers)
        if not r.ok:
            return None, None, get_response_elapsed(r)

        buffer = r.content
        mime_type = get_buffer_mime_type(buffer)
        if not mime_type.startswith('image'):
            return None, None, get_response_elapsed(r)

        return buffer, r.headers.get('ETag'), get_response_elapsed(r)

    def save_image(self, url, dir_path, name, width, height, keep_aspect_ratio=True, dominant_color=True, format='JPEG'):
        if url is None:
            return False

        # If image has already been retrieved
        # Check first if it has changed using ETag
        current_etag = None
        etag_fs_path = os.path.join(dir_path, f'{name}.etag')
        if os.path.exists(etag_fs_path):
            with open(etag_fs_path, 'r') as fp:
                current_etag = fp.read()

        # Save image file
        try:
            data, etag, _rtime = self.get_image(url, current_etag)
        except Exception:
            return False
        if data is None:
            return False

        data = convert_and_resize_image(
            data, width, height, keep_aspect_ratio=keep_aspect_ratio, dominant_color=dominant_color, format=format
        )
        if data is None:
            return False

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        fs_path = os.path.join(dir_path, f'{name}.{"jpg" if format == "JPEG" else format.lower()}')
        with open(fs_path, 'wb') as fp:
            fp.write(data)

        if etag:
            with open(etag_fs_path, 'w') as fp:
                fp.write(etag)
        elif os.path.exists(etag_fs_path):
            os.remove(etag_fs_path)

        return True

    def session_get(self, *args, **kwargs):
        try:
            r = retry_session(session=self.session).get(*args, **kwargs)
        except Exception as error:
            logger.debug(error)
            raise

        return r

    def session_patch(self, *args, **kwargs):
        try:
            r = self.session.patch(*args, **kwargs)
        except Exception as error:
            logger.debug(error)
            raise

        return r

    def session_post(self, *args, **kwargs):
        try:
            r = self.session.post(*args, **kwargs)
        except Exception as error:
            logger.debug(error)
            raise

        return r

    def session_put(self, *args, **kwargs):
        try:
            r = self.session.put(*args, **kwargs)
        except Exception as error:
            logger.debug(error)
            raise

        return r


class CustomTimeout(TimeoutSauce):
    def __init__(self, *args, **kwargs):
        if kwargs['connect'] is None:
            kwargs['connect'] = REQUESTS_TIMEOUT
        if kwargs['read'] is None:
            kwargs['read'] = REQUESTS_TIMEOUT * 2
        super().__init__(*args, **kwargs)


# Set requests timeout globally, instead of specifying ``timeout=..`` kwarg on each call
requests.adapters.TimeoutSauce = CustomTimeout


class CoverLoader(GObject.GObject):
    __gtype_name__ = 'CoverLoader'

    def __init__(self, path, pixbuf, texture, width=None, height=None, static_animation=False):
        super().__init__()

        self.path = path
        self.pixbuf = pixbuf
        self.texture = texture

        if self.pixbuf:
            self.orig_width = self.pixbuf.get_width()
            self.orig_height = self.pixbuf.get_height()
        else:
            self.orig_width = self.texture.get_width()
            self.orig_height = self.texture.get_height()

        # Compute size
        if width is None and height is None:
            self.width = self.orig_width
            self.height = self.orig_height

        elif width is None or height is None:
            ratio = self.orig_width / self.orig_height
            if width is None:
                self.width = int(height * ratio)
                self.height = height
            else:
                self.width = width
                self.height = int(width / ratio)

        else:
            self.width = width
            self.height = height

    @classmethod
    def new_from_data(cls, data, width=None, height=None, static_animation=False):
        info, mime_type = get_image_info(data)
        if not info:
            return None

        try:
            stream = Gio.MemoryInputStream.new_from_data(data, None)
            if info['is_animated'] and not static_animation:
                pixbuf = PixbufAnimation.new_from_stream(stream)
            else:
                pixbuf = Pixbuf.new_from_stream(stream)

            stream.close()
        except Exception as exc:
            # Invalid image, corrupted image, unsupported image format,...
            logger.warning(
                'Failed to create pixbuf: corrupted file or unsupported image format (%s)',
                mime_type, exc_info=exc
            )
            return None

        return cls(None, pixbuf, None, width, height, static_animation)

    @classmethod
    def new_from_file(cls, path, width=None, height=None, static_animation=False):
        info, mime_type = get_image_info(path)
        if not info:
            return None

        try:
            if info['is_animated'] and not static_animation:
                pixbuf = PixbufAnimation.new_from_file(path)
            else:
                pixbuf = Pixbuf.new_from_file(path)
        except Exception as exc:
            # Invalid image, corrupted image, unsupported image format,...
            logger.warning(
                'Failed to create pixbuf: corrupted file or unsupported image format (%s)',
                mime_type, exc_info=exc
            )
            return None

        return cls(path, pixbuf, None, width, height, static_animation)

    @classmethod
    def new_from_resource(cls, path, width=None, height=None):
        try:
            texture = Gdk.Texture.new_from_resource(path)
        except Exception:
            # Invalid image, corrupted image, unsupported image format,...
            return None

        return cls(None, None, texture, width, height, True)

    def dispose(self):
        self.pixbuf = None
        self.texture = None


class CoverPaintable(CoverLoader, Gdk.Paintable):
    __gtype_name__ = 'CoverPaintable'

    def __init__(self, path, image, texture, width=None, height=None, static_animation=False):
        CoverLoader.__init__(self, path, image, texture, width, height, static_animation)

        self.rect = Graphene.Rect().alloc()

        if isinstance(self.pixbuf, PixbufAnimation):
            self.animation_iter = self.pixbuf.get_iter(None)
            self.animation_timeout_id = None
        else:
            self.animation_iter = None
            self.animation_timeout_id = None

    def _start_animation(self):
        if not self.animation_iter or self.animation_timeout_id:
            return

        self.animation_timeout_id = GLib.timeout_add(self.animation_iter.get_delay_time(), self.on_delay)

    def _stop_animation(self):
        if not self.animation_iter or self.animation_timeout_id is None:
            return

        GLib.source_remove(self.animation_timeout_id)
        self.animation_timeout_id = None

    def dispose(self):
        self.animation_iter = None
        CoverLoader.dispose(self)

    def do_get_intrinsic_height(self):
        return self.height

    def do_get_intrinsic_width(self):
        return self.width

    def do_snapshot(self, snapshot, width, height):
        self.rect.init(0, 0, width, height)

        if self.pixbuf:
            if self.animation_iter:
                self.texture = Gdk.Texture.new_for_pixbuf(self.animation_iter.get_pixbuf())
            else:
                self.texture = Gdk.Texture.new_for_pixbuf(self.pixbuf)

        snapshot.append_texture(self.texture, self.rect)

    def on_delay(self):
        if self.animation_iter.get_delay_time() == -1:
            return GLib.SOURCE_REMOVE

        # Check if it's time to show the next frame
        if self.animation_iter.advance(None):
            self.invalidate_contents()

        return GLib.SOURCE_CONTINUE


class CoverPicture(Gtk.Picture):
    def __init__(self, paintable):
        super().__init__()
        self.set_paintable(paintable)

        if self.is_animated:
            self.connect('map', self.on_map)
            self.connect('unmap', self.on_unmap)

        self.connect('unrealize', self.on_unrealize)

    @classmethod
    def new_from_data(cls, data, width=None, height=None, static_animation=False):
        if paintable := CoverPaintable.new_from_data(data, width, height, static_animation):
            return cls(paintable)

        return None

    @classmethod
    def new_from_file(cls, path, width=None, height=None, static_animation=False):
        if paintable := CoverPaintable.new_from_file(path, width, height, static_animation):
            return cls(paintable)

        return None

    @classmethod
    def new_from_resource(cls, path, width=None, height=None):
        if paintable := CoverPaintable.new_from_resource(path, width, height):
            return cls(paintable)

        return None

    @cached_property
    def is_animated(self):
        return self.get_paintable().animation_iter is not None

    def on_map(self, _self):
        self.get_paintable()._start_animation()

    def on_unmap(self, _self):
        self.get_paintable()._stop_animation()

    def on_unrealize(self, _self):
        if self.is_animated:
            self.disconnect_by_func(self.on_map)
            self.disconnect_by_func(self.on_unmap)
            self.get_paintable()._stop_animation()

        self.disconnect_by_func(self.on_unrealize)

        self.get_paintable().dispose()


@dataclass
class ServerContent:
    """Class for describing the content available in a server (legal only)"""
    type: str = None
    license: str = None
    public_domain: bool = False  # 100% of content is in public domain



# SPDX-FileCopyrightText: 2020-2024 tijder
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: tijder
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
import datetime
import hashlib
from urllib.parse import urlparse

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
import requests

from komikku.servers import Server
from komikku.servers.utils import do_login
from komikku.utils import get_buffer_mime_type


class Izneo(Server):
    id = 'izneo'
    name = 'Izneo'
    lang = 'en'
    status = 'disabled'  # commercial platforms are no supported

    long_strip_genres = ['Webtoon', ]
    has_login = True

    base_url = 'https://www.izneo.com'
    logo_url = base_url + '/favicon.ico'
    cover_base_url = 'https://image.izneo.com'
    api_base_url = base_url + '/{0}/api/web'
    login_url = base_url + '/{0}/login'
    api_login_url = api_base_url + '/login'
    manga_url = base_url + '/{0}/{1}/{2}/{3}'
    api_manga_url = api_base_url + '/serie/{1}'
    api_library_url = api_base_url + '/library-v2'
    api_volumes_url = api_base_url + '/serie/{1}/volumes/old/{2}/{3}'
    api_others_url = api_base_url + '/serie/{1}/others/old/{2}/{3}'
    api_chapters_url = api_base_url + '/serie/{1}/chapters/old/{2}/{3}'
    api_chapter_url = base_url + '/book/{0}'
    cover_url = cover_base_url + '/{0}/images/album/{1}.jpg?v={2}'
    cover_search_url = cover_base_url + '/{0}/images/album/{1}-170or260.jpg?v={2}'

    def __init__(self, username=None, password=None):
        if username and password:
            self.do_login(username, password)

    @do_login
    def get_manga_data(self, initial_data):
        """
        Returns comic data from API

        Initial data should contain at least comic's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial_data'
        r = self.session_get(self.api_manga_url.format(self.lang, initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['name'],
            authors=[author['name'] for author in resp_data['authors']],
            scanlators=[],
            genres=[resp_data['genre']['name']],
            status='ongoing',
            chapters=[],
            synopsis=resp_data['synopsis'],
            cover=None,
            server_id=self.id,
            url=self.manga_url.format(self.lang, resp_data['shelf']['slug'], resp_data['genre']['slug'], resp_data['forUrl']),
        ))

        if cover_id := resp_data.get('cover'):
            data['cover'] = self.cover_url.format(self.lang, cover_id, resp_data.get('coverVersion', 'undefined'))
        if resp_data['isWebtoon']:
            data['genres'].append('Webtoon')

        # Volumes/Chapters/Others
        kinds = {
            'volumes': self.api_volumes_url,
            'chapters': self.api_chapters_url,
            'others': self.api_others_url,
        }
        limit = 20
        for kind, api_url in kinds.items():
            offset = 0
            while offset is not None:
                r = self.session_get(api_url.format(self.lang, initial_data['slug'], offset, limit))
                resp_data = r.json()
                if not resp_data['albums']:
                    offset = None
                    continue

                for item in resp_data['albums']:
                    if not item['inUserLibrary']:
                        continue

                    title = item['title']
                    if item['volume']:
                        title = '{0} - {1}'.format(item['volume'], title)

                    data['chapters'].append(dict(
                        slug=item['id'],
                        title=title,
                        date=datetime.datetime.fromtimestamp(item['version']).date(),
                    ))

                offset += limit

        return data

    @do_login
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        resp_data = r.json()
        if resp_data['status'] != 'ok':
            return None

        data = dict(
            pages=[],
        )
        for page in r.json()['data']['pages']:
            data['pages'].append(dict(
                slug=None,
                image=page['src'],
                key=page['key'],
                iv=page['iv'],
            ))

        return data

    @do_login
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Origin': self.base_url,
                'Referer': self.base_url + '/',
            }
        )
        if r.status_code != 200:
            return None

        cipher = Cipher(algorithms.AES(base64.b64decode(page['key'])), modes.CBC(base64.b64decode(page['iv'])))
        decryptor = cipher.decryptor()
        content = decryptor.update(r.content) + decryptor.finalize()
        mime_type = get_buffer_mime_type(content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=content,
            mime_type=mime_type,
            name=page['image'].split('?')[0].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return url

    @do_login
    def get_most_populars(self):
        """
        Returns all comics available in user's collection
        """
        return self.search('')

    def login(self, username, password):
        if not username or not password:
            return False

        cookie = requests.cookies.create_cookie(
            name='lang',
            value=self.lang,
            domain=urlparse(self.base_url).netloc,
            path='/',
            expires=(datetime.datetime.now() + datetime.timedelta(days=365)).timestamp(),
        )
        self.session.cookies.set_cookie(cookie)

        r = self.session_get(self.login_url.format(self.lang))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        r = self.session_post(
            self.api_login_url.format(self.lang),
            json={
                'login': username,
                'password': hashlib.md5(password.encode()).hexdigest(),
                'rememberMe': True,
            },
            headers={
                'Origin': self.base_url,
                'Referer': self.login_url.format(self.lang),
            }
        )

        self.save_session()

        return True

    @do_login
    def search(self, term):
        """
        Searches in user's collection
        """
        results = []
        offset = 0

        while offset is not None:
            r = self.session_post(
                self.api_library_url.format(self.lang),
                json={
                    'search': term,
                    'itemType': 'series',
                    'order': 'order-date',
                    'offset': offset,
                    'limit': 30,
                    'collection': None,
                    'genre': None,
                    'shelf': None,
                    'hide-marked-as-read': None,
                    'show-removed': False,
                },
                headers={
                    'Origin': self.base_url,
                    'Referer': self.base_url + '/{0}/bibliotheque'.format(self.lang),
                }
            )
            if r.status_code != 200:
                return None

            try:
                resp_data = r.json()
            except Exception:
                return None

            if not resp_data['totalSeries']:
                break

            for serie in resp_data['series']:
                results.append(dict(
                    name=serie['name'],
                    slug=serie['id'],
                    cover=self.cover_search_url.format(self.lang, serie['coverEan'], serie['version']),
                ))

            offset += resp_data['limit'][1]
            if offset > resp_data['totalSeries']:
                # No more comics to load
                offset = None

        return results


class Izneo_de(Izneo):
    id = 'izneo_de'
    name = 'Izneo'
    lang = 'de'


class Izneo_fr(Izneo):
    id = 'izneo_fr'
    name = 'Izneo'
    lang = 'fr'


class Yieha(Izneo):
    id = 'yieha:izneo'
    name = 'Yieha'
    lang = 'nl'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Mangalek(Madara):
    id = 'mangalek'
    name = 'مانجا ليك Mangalek'
    lang = 'ar'

    has_cf = True

    date_format = '%Y ,%d %B'

    base_url = 'https://lek-manga.net'
    logo_url = 'https://io.lek-manga.net/wp-content/app/lekmanganet/512.png'
    chapter_url = base_url + '/manga/{0}/{1}/'

    bypass_cf_url = base_url + '/manga/apotheosis/'

    # Mirrors
    # https://lekmanga.online/
    # https://lekmanga.site
    # https://like-manga.net
    # https://manga-leko.org



# SPDX-FileCopyrightText: 2022-2024 CakesTwix
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: CakesTwix <oleg.kiryazov@gmail.com>

import logging
import re
import time
from urllib.parse import urlparse

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge

logger = logging.getLogger('komikku.servers.remanga')

re_tags_remove = re.compile(r'<[^>]+>')


class Remanga(Server):
    id = 'remanga'
    name = 'Remanga'
    lang = 'ru'

    has_captcha = True

    base_url = 'https://xn--80aaig9ahr.xn--c1avg'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'
    api_base_url = 'https://api.xn--80aaig9ahr.xn--c1avg'
    api_search_url = api_base_url + '/api/search/'
    api_manga_list_url = api_base_url + '/api/search/catalog/'
    api_manga_url = api_base_url + '/api/titles/{0}/'
    api_chapters_url = api_base_url + '/api/titles/chapters/'

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'content-type': 'application/json',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['content']

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        data['name'] = resp_data['rus_name']
        data['cover'] = self.base_url + resp_data['img']['high']

        # Details
        data['scanlators'] = [publisher['name'] for publisher in resp_data['publishers']]
        data['genres'] = [genre['name'] for genre in resp_data['genres']]

        if resp_data['status']['id'] == 1:
            data['status'] = 'ongoing'
        elif resp_data['status']['id'] == 0:
            data['status'] = 'complete'
        elif resp_data['status']['id'] == 2:
            data['status'] = 'suspended'

        # Synopsis
        data['synopsis'] = re_tags_remove.sub('', resp_data['description'])

        # Chapters
        chapters = []
        branch_id = resp_data['branches'][0]['id']
        page = 1
        while True:
            start = time.time()
            r = self.session_get(
                self.api_chapters_url,
                params=dict(
                    branch_id=branch_id,
                    ordering='-index',
                    user_data=1,
                    count=40,
                    page=page,
                ),
                headers={
                    'Content-Type': 'application/json',
                    'Referer': self.base_url,
                }
            )
            if r.status_code != 200:
                # Return all chapters or nothing
                return data

            resp_data = r.json()['content']
            if not resp_data:
                break

            for chapter in resp_data:
                title = '#{0}'.format(chapter['chapter'])
                if chapter['name']:
                    title = '{0} - {1}'.format(title, chapter['name'])

                chapters.append(dict(
                    slug=str(chapter['id']),  # must be a string
                    title=title,
                    num=chapter['chapter'] if is_number(chapter.get('chapter')) else None,
                    num_volume=chapter['tome'] if is_number(chapter.get('tome')) else None,
                    date=convert_date_string(chapter['upload_date'][:-10], '%Y-%m-%d'),
                ))
            page += 1

            delay = min(3 * (time.time() - start), 1)
            time.sleep(delay)

        data['chapters'] = list(reversed(chapters))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.api_chapters_url + str(chapter_slug),
            headers={
                'content-type': 'application/json',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['content']

        data = dict(
            pages=[],
        )
        for page in resp_data['pages']:
            if isinstance(page, list):
                for page_list in page:
                    data['pages'].append(dict(
                        slug=None,
                        image=page_list['link'],
                    ))
            else:
                data['pages'].append(dict(
                    slug=None,
                    image=page['link'],
                ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Host': urlparse(page['image']).netloc,
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, _url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, orderby):
        r = self.session_get(
            self.api_manga_list_url,
            params={
                'count': 50,
                'exclude_bookmarks': '',
                'ordering': '-views' if orderby == 'populars' else '-chapter_date',
                'page': 1,
            },
            headers={
                'Content-Type': 'application/json',
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['content']:
            results.append(dict(
                slug=item['dir'],
                name=item['rus_name'],
                cover=self.base_url + item['img']['mid'],
            ))

        return results

    @CompleteChallenge()
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        return self.get_manga_list(orderby='latest')

    @CompleteChallenge()
    def get_most_populars(self):
        """
        Returns most popular mangas
        """
        return self.get_manga_list(orderby='populars')

    @CompleteChallenge()
    def search(self, term):
        r = self.session_get(
            self.api_search_url,
            params={
                'query': term,
            },
            headers={
                'content-type': 'application/json',
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['content']:
            results.append(dict(
                slug=item['dir'],
                name=item['rus_name'],
                cover=self.base_url + item['img']['mid'],
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.iken import Iken


class Nightscans(Iken):
    id = 'nightscans'
    name = 'Qi Scans (Night scans)'
    lang = 'en'

    base_url = 'https://qimanhwa.com'
    logo_url = base_url + 'qimanhwa.ico'
    api_url = 'https://api.qimanhwa.com/api/v1'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
from gettext import ngettext
import natsort

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from komikku.models import Chapter
from komikku.models import create_db_connection
from komikku.models import Download
from komikku.models import update_rows
from komikku.utils import html_escape


class ChapterItemWrapper(GObject.Object):
    __gsignals__ = {
        'changed': (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, chapter):
        super().__init__()
        self.chapter = chapter
        self.download = None

    def emit_changed(self):
        self.emit('changed')


class ChaptersFilterListModel(Gtk.FilterListModel):
    def __init__(self, model, filter_func):
        self.filter = Gtk.CustomFilter()
        self.filter_func = filter_func
        self.filter.set_filter_func(self.filter_func)

        super().__init__(model=model, filter=self.filter)

    def invalidate_filter(self):
        self.filter.set_filter_func(self.filter_func)

    def invalidate_sort(self):
        self.get_model().sorter.set_sort_func(self.get_model().sort_func)

    def populate(self, chapters):
        self.get_model().populate(chapters)


class ChaptersSortListModel(Gtk.SortListModel):
    def __init__(self, sort_func):
        self.sorter = Gtk.CustomSorter()
        self.sort_func = sort_func
        self.sorter.set_sort_func(self.sort_func)

        super().__init__(model=Gio.ListStore(item_type=ChapterItemWrapper), sorter=self.sorter)

    def invalidate_sort(self):
        self.sorter.set_sort_func(self.sort_func)

    def populate(self, chapters):
        items = []
        for chapter in chapters:
            items.append(ChapterItemWrapper(chapter))

        self.get_model().splice(0, self.get_model().props.n_items, items)


class ChaptersList:
    def __init__(self, card):
        self.card = card

        self.selection_last_selected_position = None
        self.selection_mode_range = False
        self.selection_positions = []
        self.selection_click_position = None

        self.factory = Gtk.SignalListItemFactory()
        self.factory.connect('bind', self.on_factory_bind)
        self.factory.connect('setup', self.on_factory_setup)
        self.factory.connect('unbind', self.on_factory_unbind)

        self.list_model = ChaptersFilterListModel(ChaptersSortListModel(self.sort_func), self.filter_func)
        self.model = Gtk.MultiSelection.new(self.list_model)
        self.selection_changed_handler_id = self.model.connect('selection-changed', self.on_selection_changed)

        self.listview = self.card.chapters_listview
        # Remove unwanted style class 'view' which changes background color in dark appearance!
        self.listview.remove_css_class('view')
        self.listview.set_factory(self.factory)
        self.listview.set_model(self.model)
        self.listview.set_show_separators(True)
        self.listview.set_single_click_activate(True)
        self.listview.connect('activate', self.on_row_activate)

        # Chapters selection mode ActionBar
        self.chapters_selection_mode_actionbar = self.card.chapters_selection_mode_actionbar
        self.card.chapters_selection_mode_menubutton.set_menu_model(self.card.builder.get_object('menu-card-selection-mode'))

        # Gesture to detect long press on mouse button 1 and enter in selection mode
        self.gesture_long_press = Gtk.GestureLongPress.new()
        self.gesture_long_press.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture_long_press.set_touch_only(False)
        self.gesture_long_press.connect('pressed', self.on_long_press)
        self.listview.add_controller(self.gesture_long_press)

        self.card.window.downloader.connect('download-changed', self.update_chapter_item)

    @property
    def sort_order(self):
        return self.card.manga.sort_order or 'desc'

    def add_actions(self):
        # Menu actions in selection mode
        download_selected_chapters_action = Gio.SimpleAction.new('card.download-selected-chapters', None)
        download_selected_chapters_action.connect('activate', self.download_selected_chapters)
        self.card.window.application.add_action(download_selected_chapters_action)

        mark_selected_chapters_as_read_action = Gio.SimpleAction.new('card.mark-selected-chapters-read', None)
        mark_selected_chapters_as_read_action.connect('activate', self.toggle_selected_chapters_read_status, 1)
        self.card.window.application.add_action(mark_selected_chapters_as_read_action)

        mark_selected_chapters_as_unread_action = Gio.SimpleAction.new('card.mark-selected-chapters-unread', None)
        mark_selected_chapters_as_unread_action.connect('activate', self.toggle_selected_chapters_read_status, 0)
        self.card.window.application.add_action(mark_selected_chapters_as_unread_action)

        clear_selected_chapters_action = Gio.SimpleAction.new('card.clear-selected-chapters', None)
        clear_selected_chapters_action.connect('activate', self.clear_selected_chapters, False)
        self.card.window.application.add_action(clear_selected_chapters_action)

        clear_reset_selected_chapters_action = Gio.SimpleAction.new('card.clear-reset-selected-chapters', None)
        clear_reset_selected_chapters_action.connect('activate', self.clear_selected_chapters, True)
        self.card.window.application.add_action(clear_reset_selected_chapters_action)

        select_all_chapters_action = Gio.SimpleAction.new('card.select-all-chapters', None)
        select_all_chapters_action.connect('activate', self.select_all)
        self.card.window.application.add_action(select_all_chapters_action)

        # Chapters menu actions
        download_chapter_action = Gio.SimpleAction.new('card.download-chapter', GLib.VariantType.new('q'))
        download_chapter_action.connect('activate', self.download_chapter)
        self.card.window.application.add_action(download_chapter_action)

        clear_chapter_action = Gio.SimpleAction.new('card.clear-chapter', GLib.VariantType.new('q'))
        clear_chapter_action.connect('activate', self.clear_chapter, False)
        self.card.window.application.add_action(clear_chapter_action)

        clear_reset_chapter_action = Gio.SimpleAction.new('card.clear-reset-chapter', GLib.VariantType.new('q'))
        clear_reset_chapter_action.connect('activate', self.clear_chapter, True)
        self.card.window.application.add_action(clear_reset_chapter_action)

        mark_chapter_as_read_action = Gio.SimpleAction.new('card.mark-chapter-read', GLib.VariantType.new('q'))
        mark_chapter_as_read_action.connect('activate', self.toggle_chapter_read_status, 1)
        self.card.window.application.add_action(mark_chapter_as_read_action)

        mark_chapter_as_unread_action = Gio.SimpleAction.new('card.mark-chapter-unread', GLib.VariantType.new('q'))
        mark_chapter_as_unread_action.connect('activate', self.toggle_chapter_read_status, 0)
        self.card.window.application.add_action(mark_chapter_as_unread_action)

        mark_previous_chapters_as_read_action = Gio.SimpleAction.new('card.mark-previous-chapters-read', GLib.VariantType.new('q'))
        mark_previous_chapters_as_read_action.connect('activate', self.set_previous_chapters_as_read)
        self.card.window.application.add_action(mark_previous_chapters_as_read_action)

    def clear_chapter(self, _action, position, reset):
        # Clear and reset chapter
        item = self.list_model.get_item(position.get_uint16())
        if Download.get_by_chapter_id(item.chapter.id) is not None:
            # Prevent reset of a chapter that is currently downloaded or scheduled for download
            return

        item.chapter.clear(reset=reset)
        item.chapter = Chapter.get(item.chapter.id)
        item.emit_changed()

        self.card.refresh(unread_chapters=reset, info=True)
        self.card.window.library.refresh_on_manga_state_changed(self.card.manga)

    def clear_selected_chapters(self, _action, _param, reset):
        # Clear and reset selected chapters
        db_conn = create_db_connection()

        items = self.get_selected_chapters_items()
        Chapter.clear_many([item.chapter for item in items], reset=reset)
        for item in items:
            item.chapter = Chapter.get(item.chapter.id, db_conn=db_conn)
            item.emit_changed()

        self.card.refresh(unread_chapters=reset, info=True)
        self.card.window.library.refresh_on_manga_state_changed(self.card.manga)

        self.card.leave_selection_mode()

    def download_chapter(self, _action, position):
        item = self.list_model.get_item(position.get_uint16())
        self.card.window.downloader.add([item.chapter], emit_signal=True)
        self.card.window.downloader.start()
        item.emit_changed()

    def download_selected_chapters(self, _action, _gparam):
        self.card.window.downloader.add([item.chapter for item in self.get_selected_chapters_items()], emit_signal=True)
        self.card.window.downloader.start()

        self.card.leave_selection_mode()

    def enter_selection_mode(self, init=False):
        self.chapters_selection_mode_actionbar.set_revealed(True)
        self.listview.set_single_click_activate(False)

        if init:
            # Init selection with clicked row (stored in self.selection_click_position)
            self.on_selection_changed(None, None, None)

    def filter_func(self, item: ChapterItemWrapper) -> int:
        if not self.card.manga.filters or not self.card.manga.filters.get('scanlators'):
            # No scanlators filter
            return True

        if not item.chapter.scanlators:
            # Chapter has no scanlators defined
            # It is not filtered if 'Unknown' virtual scanlator does not belong to filter
            return 'Unknown' not in self.card.manga.filters['scanlators']

        return len(set(item.chapter.scanlators or []).difference(set(self.card.manga.filters['scanlators']))) > 0

    def get_selected_chapters_items(self):
        items = []

        bitsec = self.model.get_selection()
        for index in range(bitsec.get_size()):
            position = bitsec.get_nth(index)
            if self.sort_order.endswith('desc'):
                items.insert(0, self.list_model.get_item(position))
            else:
                items.append(self.list_model.get_item(position))

        return items

    def leave_selection_mode(self):
        self.model.unselect_all()

        self.selection_last_selected_position = None
        self.selection_mode_range = False
        self.selection_positions = []

        self.listview.set_single_click_activate(True)
        self.chapters_selection_mode_actionbar.set_revealed(False)

    def on_factory_bind(self, _factory: Gtk.ListItemFactory, list_item: Gtk.ListItem):
        row = list_item.get_child()
        row.connect_signals()
        row.populate(list_item.get_item())

    def on_factory_setup(self, _factory: Gtk.ListItemFactory, list_item: Gtk.ListItem):
        list_item.set_child(ChaptersListRow(self.card))

    def on_factory_unbind(self, _factory: Gtk.ListItemFactory, list_item: Gtk.ListItem):
        list_item.get_child().disconnect_signals()

    def on_long_press(self, _controller, _x, _y):
        if not self.card.selection_mode:
            self.card.enter_selection_mode(init=True)
        elif not self.selection_mode_range:
            self.selection_mode_range = True
            self.on_selection_changed(None, None, None)

    def on_row_activate(self, _listview, position):
        # Prevent unwanted activations: double-click in selection mode
        if self.card.selection_mode:
            return

        chapter = self.list_model.get_item(position).chapter
        self.card.window.reader.init(self.card.manga, chapter)

    def on_selection_changed(self, model, _position, _n_items):
        if not self.card.selection_mode:
            return

        # Here we try to allow multiple selection.
        # A short click selects or unselects a row.
        # A long press selects a range of rows (selection_last_selected_position => selection_click_position).
        # When a row is clicked, selection is lost so it must have been saved previously.
        # We build the new selection from the previous one +/- the clicked row or the previous one + a range of rows.
        if self.selection_click_position is not None:
            click_position = self.selection_click_position

            selected = Gtk.Bitset.new_empty()
            mask = Gtk.Bitset.new_empty()

            if not self.selection_mode_range:
                # Single row selection/unselection
                self.selection_click_position = None

                mask.add(click_position)
                if click_position not in self.selection_positions:
                    self.selection_positions.append(click_position)
                    self.selection_last_selected_position = click_position
                else:
                    self.selection_positions.remove(click_position)

                for position in self.selection_positions:
                    mask.add(position)
                    selected.add(position)
            else:
                # Range selection
                for position in self.selection_positions:
                    mask.add(position)

                if self.selection_last_selected_position < click_position:
                    for position in range(self.selection_last_selected_position, click_position + 1):
                        mask.add(position)
                else:
                    for position in range(click_position, self.selection_last_selected_position + 1):
                        mask.add(position)

                selected = mask

                # In selection mode range, event are emitted twice:
                # - First is a fake event `emitted` in on_long_press method
                # - Second is emitted when left click is released (in this case, model is not None)
                # State must be maintained until the second one because selection must be replayed
                if model:
                    self.selection_click_position = None
                    self.selection_last_selected_position = click_position
                    self.selection_mode_range = False

            with self.model.handler_block(self.selection_changed_handler_id):
                self.model.set_selection(selected, mask)

        self.selection_positions = []
        bitsec = self.model.get_selection()
        for index in range(bitsec.get_size()):
            self.selection_positions.append(bitsec.get_nth(index))

        number = len(self.selection_positions)
        if number:
            self.card.title.set_subtitle(ngettext('{0} selected', '{0} selected', number).format(number))
        else:
            self.card.leave_selection_mode()

    def on_sort_order_changed(self, _action, variant):
        value = variant.get_string()
        if value == self.card.manga.sort_order:
            return

        self.card.manga.update(dict(sort_order=value))
        self.set_sort_order()

    def populate(self):
        chapters = self.card.manga.chapters
        self.list_model.populate(chapters)

        if len(chapters) > 0:
            self.card.toggle_resume(True)
        else:
            self.card.window.add_notification(_('No chapters found'))
            self.card.toggle_resume(False)

    def refresh(self, chapters):
        for chapter in chapters:
            self.update_chapter_item(chapter=chapter)

    def select_all(self, *args):
        if self.card.stack.get_visible_child_name() != 'chapters':
            return

        if not self.card.selection_mode:
            self.card.enter_selection_mode()

        self.model.select_all()

    def set_previous_chapters_as_read(self, _action, position):
        chapters_ids = []
        chapters_data = []

        self.card.window.activity_indicator.set_visible(True)

        item = self.list_model.get_item(position.get_uint16())
        rank = item.chapter.rank

        # First, update DB
        for item in self.list_model:
            chapter = item.chapter
            if chapter.rank >= rank:
                continue

            chapters_ids.append(chapter.id)
            chapters_data.append(dict(
                last_page_read_index=None,
                read_progress=None,
                read=True,
                recent=False,
            ))

        db_conn = create_db_connection()

        with db_conn:
            res = update_rows(db_conn, 'chapters', chapters_ids, chapters_data)

        if res:
            # Then, if DB update succeeded, update chapters rows
            for item in self.list_model:
                chapter = item.chapter
                if chapter.rank >= rank:
                    continue

                chapter.last_page_read_index = None
                chapter.read = True
                chapter.recent = False

                item.emit_changed()

            self.card.window.activity_indicator.set_visible(False)

            self.card.refresh(unread_chapters=True)
            self.card.window.library.refresh_on_manga_state_changed(self.card.manga)
        else:
            self.card.window.activity_indicator.set_visible(False)
            self.card.window.add_notification(_('Failed to update chapters reading status'))

    def set_sort_order(self, invalidate=True):
        self.card.sort_order_action.set_state(GLib.Variant('s', self.sort_order))
        if invalidate:
            self.list_model.invalidate_sort()

    def sort_func(self, item1: ChapterItemWrapper, item2: ChapterItemWrapper, *args) -> int:
        """
        This function gets two children and has to return:
        - a negative integer if the firstone should come before the second one
        - zero if they are equal
        - a positive integer if the second one should come before the first one
        """
        if self.sort_order in ('asc', 'desc'):
            if item1.chapter.rank > item2.chapter.rank:
                return -1 if self.sort_order == 'desc' else 1

            if item1.chapter.rank < item2.chapter.rank:
                return 1 if self.sort_order == 'desc' else -1

        elif self.sort_order in ('date-asc', 'date-desc') and item1.chapter.date and item2.chapter.date:
            if item1.chapter.date > item2.chapter.date and item1.chapter.id > item2.chapter.id:
                return -1 if self.sort_order == 'date-desc' else 1

            if item1.chapter.date < item2.chapter.date and item1.chapter.id < item2.chapter.id:
                return 1 if self.sort_order == 'date-desc' else -1

        elif self.sort_order in ('natural-asc', 'natural-desc'):
            lst = natsort.natsorted([item1.chapter.title, item2.chapter.title], alg=natsort.ns.INT | natsort.ns.IC)
            if lst[0] == item1.chapter.title:
                return 1 if self.sort_order == 'natural-desc' else -1

            return -1 if self.sort_order == 'natural-desc' else 1

        return 0

    def toggle_chapter_read_status(self, _action, position, read):
        item = self.list_model.get_item(position.get_uint16())
        chapter = item.chapter

        data = dict(
            last_page_read_index=None,
            read_progress=None,
            read=read,
            recent=False,
        )
        chapter.update(data)

        item.emit_changed()

        self.card.refresh(unread_chapters=True)
        self.card.window.library.refresh_on_manga_state_changed(self.card.manga)

    def toggle_selected_chapters_read_status(self, _action, _gparam, read):
        chapters_ids = []
        chapters_data = []

        self.card.window.activity_indicator.set_visible(True)

        # First, update DB
        for item in self.get_selected_chapters_items():
            chapters_ids.append(item.chapter.id)
            chapters_data.append(dict(
                last_page_read_index=None,
                read_progress=None,
                read=read,
                recent=False,
            ))

        db_conn = create_db_connection()

        with db_conn:
            res = update_rows(db_conn, 'chapters', chapters_ids, chapters_data)

        if res:
            # Then, if DB update succeeded, update chapters rows
            def update_chapters_rows():
                db_conn = create_db_connection()

                for item in self.get_selected_chapters_items():
                    item.chapter = Chapter.get(item.chapter.id, db_conn=db_conn)
                    item.emit_changed()
                    yield True

                self.card.leave_selection_mode()
                self.card.window.activity_indicator.set_visible(False)

                self.card.refresh(unread_chapters=True)
                self.card.window.library.refresh_on_manga_state_changed(self.card.manga)

            def run_generator(func):
                gen = func()
                GLib.idle_add(lambda: next(gen, False), priority=GLib.PRIORITY_DEFAULT_IDLE)

            run_generator(update_chapters_rows)
        else:
            self.card.window.activity_indicator.set_visible(False)
            self.card.leave_selection_mode()
            self.card.window.add_notification(_('Failed to update chapters reading status'))

    def update_chapter_item(self, _downloader=None, download=None, chapter=None):
        """
        Update a specific chapter row
        - used when download status change (via signal from Downloader)
        - used when we come back from reader to update last page read
        """
        if chapter is None:
            chapter = download.chapter

        if self.card.window.page not in ('card', 'reader') or self.card.manga.id != chapter.manga_id:
            return

        position = -1
        for item in self.list_model:
            position += 1
            if item.chapter.id == chapter.id:
                item.chapter = chapter
                item.download = download
                item.emit_changed()
                break

        self.card.window.library.refresh_on_manga_state_changed(self.card.manga)


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/card_chapters_list_row.ui')
class ChaptersListRow(Gtk.Box):
    __gtype_name__ = 'ChaptersListRow'

    primary_hbox = Gtk.Template.Child('primary_hbox')
    title_label = Gtk.Template.Child('title_label')
    scanlators_label = Gtk.Template.Child('scanlators_label')
    menubutton = Gtk.Template.Child('menubutton')
    secondary_hbox = Gtk.Template.Child('secondary_hbox')
    badge_label = Gtk.Template.Child('badge_label')
    subtitle_label = Gtk.Template.Child('subtitle_label')
    download_progress_progressbar = Gtk.Template.Child('download_progress_progressbar')
    download_stop_button = Gtk.Template.Child('download_stop_button')
    read_progress_label = Gtk.Template.Child('read_progress_label')

    def __init__(self, card):
        Gtk.Box.__init__(self)

        self.card = card
        self.chapter = None

        # Menu button
        self.menubutton.set_menu_model(Gio.Menu())

        # Gesture to detect click
        self.gesture_click = Gtk.GestureClick.new()
        self.gesture_click.set_button(0)
        self.add_controller(self.gesture_click)

    @property
    def position(self):
        position = -1
        for item in self.card.chapters_list.list_model:
            position += 1
            if item.chapter.id == self.chapter.id:
                break

        return position

    def connect_signals(self):
        self.popover_show_handler_id = self.menubutton.get_popover().connect('show', self.update_menu)
        self.gesture_click_handler_id = self.gesture_click.connect('pressed', self.on_button_clicked)

    def disconnect_signals(self):
        self.menubutton.get_popover().disconnect(self.popover_show_handler_id)
        self.gesture_click.disconnect(self.gesture_click_handler_id)
        self.download_stop_button.disconnect(self.download_stop_button_clicked_handler_id)

    def on_button_clicked(self, _gesture, _n_press, _x, _y):
        button = self.gesture_click.get_current_button()
        if button == 1:
            # Left button
            # Store row position
            self.card.chapters_list.selection_click_position = self.position
            return Gdk.EVENT_STOP

        if button == 3 and not self.card.selection_mode:
            # Right button
            # Store row position and enter selection mode
            self.card.chapters_list.selection_click_position = self.position
            self.card.enter_selection_mode(init=True)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def populate(self, item: ChapterItemWrapper, update=False):
        self.chapter = item.chapter

        # Connect events
        if not update:
            item.connect('changed', self.populate, True)
            self.download_stop_button_clicked_handler_id = self.download_stop_button.connect(
                'clicked', lambda _button, chapter: self.card.window.downloader.cancel([chapter]), self.chapter
            )

        self.title_label.set_label(self.chapter.title)
        self.title_label.remove_css_class('dimmed')
        self.title_label.remove_css_class('warning')
        if self.chapter.read:
            # Chapter reading ended
            self.title_label.add_css_class('dimmed')
        elif self.chapter.last_page_read_index is not None:
            # Chapter reading started
            self.title_label.add_css_class('warning')

        if self.chapter.scanlators:
            self.scanlators_label.set_markup(html_escape(', '.join(self.chapter.scanlators)))
            self.scanlators_label.set_visible(True)
        else:
            self.scanlators_label.set_text('')
            self.scanlators_label.set_visible(False)

        #
        # Recent badge, date, download status, page counter
        #
        show_secondary_hbox = False

        if self.chapter.recent == 1:
            self.badge_label.set_visible(True)
            show_secondary_hbox = True
        else:
            self.badge_label.set_visible(False)

        # Date + Download status (text or progress bar)
        download_status = None
        if self.chapter.downloaded:
            download_status = 'downloaded'
        else:
            if item.download is None:
                item.download = Download.get_by_chapter_id(self.chapter.id)
            if item.download:
                download_status = item.download.status

        text = [self.chapter.date.strftime(_('%m/%d/%Y'))] if self.chapter.date else []
        if download_status is not None and download_status != 'downloading' and not self.card.manga.is_local:
            text.append(_(Download.STATUSES[download_status]).upper())

        self.subtitle_label.set_text(' · '.join(text))
        if text:
            show_secondary_hbox = True

        if download_status == 'downloading':
            show_secondary_hbox = True
            self.subtitle_label.set_hexpand(False)
            self.download_progress_progressbar.set_visible(True)
            self.download_stop_button.set_visible(True)
            self.read_progress_label.set_visible(False)

            # Set download progress
            self.download_progress_progressbar.set_fraction(item.download.percent / 100)
        else:
            self.subtitle_label.set_hexpand(True)
            self.download_progress_progressbar.set_visible(False)
            self.download_stop_button.set_visible(False)

            # Read progress: nb read / nb pages
            if not self.chapter.read:
                if self.chapter.last_page_read_index is not None:
                    show_secondary_hbox = True

                    # Nb read / nb pages
                    nb_pages = len(self.chapter.pages) if self.chapter.pages else '?'
                    self.read_progress_label.set_text(f'{self.chapter.last_page_read_index + 1}/{nb_pages}')
                    self.read_progress_label.set_visible(True)
                elif text:
                    self.read_progress_label.set_visible(False)
            elif text:
                self.read_progress_label.set_visible(False)

        if show_secondary_hbox:
            self.secondary_hbox.set_visible(True)
            self.primary_hbox.props.margin_top = 6
            self.primary_hbox.props.margin_bottom = 6
        else:
            self.secondary_hbox.set_visible(False)
            # Increase top and bottom margins so that all rows have the same height
            self.primary_hbox.props.margin_top = 17
            self.primary_hbox.props.margin_bottom = 16

    def update_menu(self, popover):
        if self.card.selection_mode:
            # Prevent popover to pop up in selection mode
            popover.popdown()
            return

        position = self.position
        menu_model = self.menubutton.get_menu_model()
        menu_model.remove_all()

        if not self.chapter.downloaded and not self.card.manga.is_local:
            section_menu_model = Gio.Menu()
            menu_item = Gio.MenuItem.new(_('Download'))
            menu_item.set_action_and_target_value('app.card.download-chapter', GLib.Variant.new_uint16(position))
            section_menu_model.append_item(menu_item)

            section = Gio.MenuItem.new_section(None, section_menu_model)
            menu_model.append_item(section)

        section_menu_model = Gio.Menu()

        if self.chapter.clearable:
            menu_item = Gio.MenuItem.new(_('Clear'))
            menu_item.set_action_and_target_value('app.card.clear-chapter', GLib.Variant.new_uint16(position))
            section_menu_model.append_item(menu_item)

        menu_item = Gio.MenuItem.new(_('Clear and Reset') if self.chapter.clearable else _('Reset'))
        menu_item.set_action_and_target_value('app.card.clear-reset-chapter', GLib.Variant.new_uint16(position))
        section_menu_model.append_item(menu_item)

        if section_menu_model.get_n_items():
            section = Gio.MenuItem.new_section(None, section_menu_model)
            menu_model.append_item(section)

        section_menu_model = Gio.Menu()
        if not self.chapter.read:
            menu_item = Gio.MenuItem.new(_('Mark as Read'))
            menu_item.set_action_and_target_value('app.card.mark-chapter-read', GLib.Variant.new_uint16(position))
            section_menu_model.append_item(menu_item)
        if self.chapter.read or self.chapter.last_page_read_index is not None:
            menu_item = Gio.MenuItem.new(_('Mark as Unread'))
            menu_item.set_action_and_target_value('app.card.mark-chapter-unread', GLib.Variant.new_uint16(position))
            section_menu_model.append_item(menu_item)

        menu_item = Gio.MenuItem.new(_('Mark Previous Chapters as Read'))
        menu_item.set_action_and_target_value('app.card.mark-previous-chapters-read', GLib.Variant.new_uint16(position))
        section_menu_model.append_item(menu_item)

        section = Gio.MenuItem.new_section(None, section_menu_model)
        menu_model.append_item(section)



# SPDX-FileCopyrightText: 2025 gondolyr <gondolyr+code@posteo.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# Author: gondolyr <gondolyr+code@posteo.org>

#
# API doc: https://api.comick.fun/docs/
#

from gettext import gettext as _
import logging

try:
    # For some reasons, under flatpak sandbox, API calls return 403 errors!
    # API works perfectly well in terminal with curl (Note that `--tlsv1.3` option is required with GNOME 48 runtime)
    # Only solution found, use curl_cffi (JA3/TLS and HTTP2 fingerprints impersonation) in place of requests
    # What's wrong with sandbox?
    from curl_cffi import requests
except Exception:
    # Server will be disabled
    requests = None

from komikku.consts import REQUESTS_TIMEOUT
from komikku.servers import Server
from komikku.servers.exceptions import NotFoundError
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type

logger = logging.getLogger('komikku.servers.comick')

CHAPTERS_PER_REQUEST = 100
SEARCH_RESULTS_LIMIT = 50

SERVER_NAME = 'ComicK'


class Comick(Server):
    id = 'comick'
    name = SERVER_NAME
    lang = 'en'
    lang_code = 'en'

    is_nsfw = True
    # status = 'enabled' if requests is not None else 'disabled'
    status = 'disabled'  # Shut down 09/2025

    base_url = 'https://comick.io'
    logo_url = base_url + '/favicon.ico'

    api_base_url = 'https://api.comick.fun'
    api_search_url = api_base_url + '/v1.0/search/'
    api_latest_chapters_url = api_base_url + '/chapter/'
    api_manga_url = api_base_url + '/comic/{slug}/'
    api_manga_chapters_url = api_base_url + '/comic/{hid}/chapters'
    api_chapter_url = api_base_url + '/chapter/{hid}/'

    manga_url = base_url + '/comic/{slug}'
    image_url = 'https://meo.comick.pictures/{b2key}'

    filters = [
        {
            'key': 'ratings',
            'type': 'select',
            'name': _('Rating'),
            'description': _('Filter by Content Ratings'),
            'value_type': 'multiple',
            'options': [
                {'key': 'safe', 'name': _('Safe'), 'default': True},
                {'key': 'suggestive', 'name': _('Suggestive'), 'default': True},
                {'key': 'erotica', 'name': _('Erotica'), 'default': False},
            ],
        },
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': '1', 'name': _('Ongoing'), 'default': False},
                {'key': '2', 'name': _('Completed'), 'default': False},
                {'key': '3', 'name': _('Canceled'), 'default': False},
                {'key': '4', 'name': _('Paused'), 'default': False},
            ],
        },
        {
            'key': 'origination',
            'type': 'select',
            'name': _('Origination'),
            'description': _('[Latest Updates only] Filter by Origination'),
            'value_type': 'multiple',
            'options': [
                {'key': 'manga', 'name': _('Manga'), 'default': False},
                {'key': 'manhwa', 'name': _('Manhwa'), 'default': False},
                {'key': 'manhua', 'name': _('Manhua'), 'default': False},
            ],
        },
        {
            'key': 'countries',
            'type': 'select',
            'name': _('Countries'),
            'description': _(
                '[Search and Most Popular only] Filter by Country (origination)'
            ),
            'value_type': 'multiple',
            'options': [
                {'key': 'jp', 'name': _('Japan (Manga)'), 'default': False},
                {'key': 'kr', 'name': _('Korea (Manhwa)'), 'default': False},
                {'key': 'cn', 'name': _('China (Manhua)'), 'default': False},
            ],
        },
        {
            'key': 'publication_demographics',
            'type': 'select',
            'name': _('Publication Demographic'),
            'description': _('Filter by Publication Demographics'),
            'value_type': 'multiple',
            'options': [
                {'key': '1', 'name': _('Shounen'), 'default': False},
                {'key': '2', 'name': _('Shoujo'), 'default': False},
                {'key': '3', 'name': _('Seinen'), 'default': False},
                {'key': '4', 'name': _('Josei'), 'default': False},
                {'key': '5', 'name': _('None'), 'default': False},
            ],
        },
        {
            'key': 'tags',
            'type': 'select',
            'name': _('Tags'),
            'description': _('Filter by Formats'),
            'value_type': 'multiple',
            'options': [
                {'key': '4-koma', 'name': _('4-Koma'), 'default': False},
                {'key': 'adaptation', 'name': _('Adaptation'), 'default': False},
                {'key': 'anthology', 'name': _('Anthology'), 'default': False},
                {'key': 'award-winning', 'name': _('Award Winning'), 'default': False},
                {'key': 'doujinshi', 'name': _('Doujinshi'), 'default': False},
                {'key': 'fan-colored', 'name': _('Fan Colored'), 'default': False},
                {'key': 'full-color', 'name': _('Full Color'), 'default': False},
                {'key': 'long-strip', 'name': _('Long Strip'), 'default': False},
                {
                    'key': 'official-colored',
                    'name': _('Official Colored'),
                    'default': False,
                },
                {'key': 'oneshot', 'name': _('Oneshot'), 'default': False},
                {'key': 'user-created', 'name': _('User Created'), 'default': False},
                {'key': 'web-comic', 'name': _('Webcomic'), 'default': False},
            ],
        },
        {
            'key': 'genres',
            'type': 'select',
            'name': _('Genres'),
            'description': _('Filter by Genres'),
            'value_type': 'multiple',
            'options': [
                {'key': 'action', 'name': _('Action'), 'default': False},
                {'key': 'adult', 'name': _('Adult'), 'default': False},
                {'key': 'adventure', 'name': _('Adventure'), 'default': False},
                {'key': 'comedy', 'name': _('Comedy'), 'default': False},
                {'key': 'crime', 'name': _('Crime'), 'default': False},
                {'key': 'drama', 'name': _('Drama'), 'default': False},
                {'key': 'ecchi', 'name': _('Ecchi'), 'default': False},
                {'key': 'fantasy', 'name': _('Fantasy'), 'default': False},
                {'key': 'gender-bender', 'name': _('Gender Bender'), 'default': False},
                {'key': 'historical', 'name': _('Historical'), 'default': False},
                {'key': 'horror', 'name': _('Horror'), 'default': False},
                {'key': 'isekai', 'name': _('Isekai'), 'default': False},
                {'key': 'magical-girls', 'name': _('Magical Girls'), 'default': False},
                {'key': 'mature', 'name': _('Mature'), 'default': False},
                {'key': 'mecha', 'name': _('Mecha'), 'default': False},
                {'key': 'medical', 'name': _('Medical'), 'default': False},
                {'key': 'mystery', 'name': _('Mystery'), 'default': False},
                {'key': 'philosophical', 'name': _('Philosophical'), 'default': False},
                {'key': 'psychological', 'name': _('Psychological'), 'default': False},
                {'key': 'romance', 'name': _('Romance'), 'default': False},
                {'key': 'sci-fi', 'name': _('Sci-Fi'), 'default': False},
                {'key': 'shoujo-ai', 'name': _('Shoujo Ai'), 'default': False},
                {'key': 'slice-of-life', 'name': _('Slice of Life'), 'default': False},
                {'key': 'sports', 'name': _('Sports'), 'default': False},
                {'key': 'superhero', 'name': _('Superhero'), 'default': False},
                {'key': 'thriller', 'name': _('Thriller'), 'default': False},
                {'key': 'tragedy', 'name': _('Tragedy'), 'default': False},
                {'key': 'wuxia', 'name': _('Wuxia'), 'default': False},
                {'key': 'yaoi', 'name': _('Yaoi'), 'default': False},
                {'key': 'yuri', 'name': _('Yuri'), 'default': False},
            ],
        },
    ]

    def __init__(self) -> None:
        if self.session is None and requests is not None:
            self.session = requests.Session(
                allow_redirects=True,
                impersonate='chrome',
                timeout=(REQUESTS_TIMEOUT, REQUESTS_TIMEOUT * 2)
            )

    def _resolve_chapters(self, comic_hid: str) -> list[dict[str, str]]:
        chapters = []
        page = 1

        while True:
            r = self.session_get(
                self.api_manga_chapters_url.format(hid=comic_hid),
                params={
                    'limit': CHAPTERS_PER_REQUEST,
                    'page': page,
                    'chap-order': 1,
                    'lang': self.lang_code,
                },
            )
            if r.status_code != 200:
                return None

            data = r.json()

            for chapter in data['chapters']:
                title = ''
                if chapter['vol']:
                    title += f'[{chapter["vol"]}] '
                if chapter['chap']:
                    title += f'#{chapter["chap"]} '
                if chapter['title']:
                    title += f'- {chapter["title"]}'

                date = chapter['publish_at'] or chapter['updated_at']

                chapters.append({
                    'slug': chapter['hid'],
                    'title': title,
                    'num': chapter['chap'],
                    'num_volume': chapter['vol'],
                    'date': convert_date_string(date.split('T')[0], format='%Y-%m-%d'),
                    'scanlators': chapter['group_name'],
                })

            if len(chapters) == data['total']:
                break

            page += 1

        return chapters

    def get_manga_data(self, initial_data: dict) -> dict:
        """
        Return manga data from the API.

        :param initial_data: Contains the following fields:
            - slug
            - name
            - cover
            - last_chapter
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(slug=initial_data['slug']),
            params={
                'tachiyomi': 'false',
            },
        )
        if r.status_code != 200:
            return None

        resp_json = r.json()

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': None,
            'cover': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
        })

        comic_data = resp_json['comic']

        data['name'] = comic_data['title']

        # Always grab the last cover.
        data['cover'] = self.image_url.format(b2key=comic_data['md_covers'][-1]['b2key'])

        data['authors'] = [author['name'] for author in resp_json['authors']]
        data['authors'] += [author['name'] for author in resp_json['artists'] if author not in data['authors']]

        data['genres'] = [genre['md_genres']['name'] for genre in comic_data['md_comic_md_genres']]

        match comic_data['status']:
            case 1:
                # Ongoing.
                data['status'] = 'ongoing'
            case 2:
                # Completed.
                data['status'] = 'complete'
            case 3:
                # Cancelled.
                data['status'] = 'suspended'
            case 4:
                # Hiatus.
                data['status'] = 'hiatus'
            case _:
                data['status'] = None

        data['synopsis'] = comic_data.get('desc')

        data['chapters'] += self._resolve_chapters(comic_data['hid'])

        return data

    def get_manga_chapter_data(
        self,
        manga_slug: str,
        manga_name: str,
        chapter_slug: str,
        chapter_url: str,
    ) -> dict:
        """
        Return manga chapter data from the API.
        """
        r = self.session_get(
            self.api_chapter_url.format(hid=chapter_slug),
            params={
                'tachiyomi': 'false',
            },
        )
        if r.status_code == 404:
            raise NotFoundError
        if r.status_code != 200:
            return None

        chapter_data = r.json()['chapter']

        title = ''
        if chapter_data['vol']:
            title += f'[{chapter_data["vol"]}] '
        if chapter_data['chap']:
            title += f'#{chapter_data["chap"]} '
        if chapter_data['title']:
            title += f'- {chapter_data["title"]}'

        pages = [
            {
                'slug': page['b2key'],
                'image': None,
            }
            for page in chapter_data['md_images']
        ]

        date = chapter_data['publish_at'] or chapter_data['updated_at']
        scanlators = chapter_data['group_name']

        return {
            'num': chapter_data['chap'],
            'num_volume': chapter_data['vol'],
            'title': title,
            'pages': pages,
            'date': convert_date_string(date.split('T')[0], format='%Y-%m-%d'),
            'scanlators': scanlators,
        }

    def get_manga_chapter_page_image(
        self, manga_slug: str, manga_name: str, chapter_slug: str, page: dict
    ) -> dict:
        """
        Return chapter page scan (image) content.
        """
        r = self.session_get(self.image_url.format(b2key=page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['slug'],
        }

    def get_manga_url(self, slug: str, url: str) -> str:
        """
        Return manga absolute URL.
        """
        return self.manga_url.format(slug=slug)

    def get_latest_updates(
        self,
        ratings: list[str] | None = None,
        statuses: list[str] | None = None,
        publication_demographics: str | None = None,
        tags: list[str] | None = None,
        genres: list[str] | None = None,
        origination: list[str] | None = None,
        countries: list[str] | None = None,
    ) -> list[dict]:
        params = {
            'lang': [self.lang_code],
            'order': 'new',
            'accept_erotic_content': 'true' if 'erotica' in ratings else 'false',
            'tachiyomi': 'false',
        }

        if origination:
            params['type'] = origination

        r = self.session_get(self.api_latest_chapters_url, params=params)
        if r.status_code != 200:
            return None

        # Use a dictionary to only have unique entries and to store the comic attributes.
        comics = {}
        for chapter in r.json():
            comic = chapter['md_comics']
            if comic['id'] in comics:
                continue

            if comic['title']:
                if comic['md_covers']:
                    cover_b2key = comic['md_covers'][-1]['b2key']
                else:
                    cover_b2key = None

                comics[comic['id']] = {
                    'slug': comic['slug'],
                    'name': comic['title'],
                    'cover': self.image_url.format(b2key=cover_b2key) if cover_b2key else None,
                    'last_chapter': comic['last_chapter'],
                }
            else:
                logger.warning('Ignoring result {}, missing name'.format(comic['id']))

        return list(comics.values())

    def get_most_populars(
        self,
        ratings: list[str] | None = None,
        statuses: list[str] | None = None,
        publication_demographics: str | None = None,
        tags: list[str] | None = None,
        genres: list[str] | None = None,
        origination: list[str] | None = None,
        countries: list[str] | None = None,
    ) -> list[dict]:
        return self.search(
            None,
            ratings=ratings,
            statuses=statuses,
            publication_demographics=publication_demographics,
            tags=tags,
            genres=genres,
            countries=countries,
            orderby='view',
        )

    def search(
        self,
        term,
        ratings: list[str] | None = None,
        statuses: list[str] | None = None,
        publication_demographics: str | None = None,
        tags: list[str] | None = None,
        genres: list[str] | None = None,
        origination: list[str] | None = None,
        countries: list[str] | None = None,
        orderby: str | None = None,
    ) -> list[dict]:
        params = {
            'genres': genres,
            'tags': tags,
            'demographic': publication_demographics,
            'limit': SEARCH_RESULTS_LIMIT,
            'tachiyomi': 'false',
        }

        if countries:
            params['country'] = countries

        if statuses:
            # The API only accepts one status.
            params['status'] = statuses[0]

        if ratings:
            # The API only accepts one content rating.
            params['content_ratings'] = ratings[0]

        if orderby:
            params['sort'] = orderby
        else:
            params['sort'] = 'view'

        if term:
            params['q'] = term

        r = self.session_get(self.api_search_url, params=params)
        if r.status_code != 200:
            return None

        results = []
        for comic in r.json():
            if comic['title']:
                if comic['md_covers']:
                    cover_b2key = comic['md_covers'][-1]['b2key']
                else:
                    cover_b2key = None

                results.append({
                    'slug': comic['slug'],
                    'name': comic['title'],
                    'cover': self.image_url.format(b2key=cover_b2key) if cover_b2key else None,
                    'last_chapter': comic['last_chapter'],
                })
            else:
                logger.warning('Ignoring result {}, missing name'.format(comic['id']))

        return results


class Comick_cs(Comick):
    id = 'comick_cs'
    name = SERVER_NAME
    lang = 'cs'
    lang_code = 'cs'


class Comick_de(Comick):
    id = 'comick_de'
    name = SERVER_NAME
    lang = 'de'
    lang_code = 'de'


class Comick_es(Comick):
    id = 'comick_es'
    name = SERVER_NAME
    lang = 'es'
    lang_code = 'es'


class Comick_es_419(Comick):
    id = 'comick_es_419'
    name = SERVER_NAME
    lang = 'es_419'
    lang_code = 'es-la'


class Comick_fr(Comick):
    id = 'comick_fr'
    name = SERVER_NAME
    lang = 'fr'
    lang_code = 'fr'


class Comick_id(Comick):
    id = 'comick_id'
    name = SERVER_NAME
    lang = 'id'
    lang_code = 'id'


class Comick_it(Comick):
    id = 'comick_it'
    name = SERVER_NAME
    lang = 'it'
    lang_code = 'it'


class Comick_ja(Comick):
    id = 'comick_ja'
    name = SERVER_NAME
    lang = 'ja'
    lang_code = 'ja'


class Comick_ko(Comick):
    id = 'comick_ko'
    name = SERVER_NAME
    lang = 'ko'
    lang_code = 'kr'


class Comick_nl(Comick):
    id = 'comick_nl'
    name = SERVER_NAME
    lang = 'nl'
    lang_code = 'nl'


class Comick_pl(Comick):
    id = 'comick_pl'
    name = SERVER_NAME
    lang = 'pl'
    lang_code = 'pl'


class Comick_pt(Comick):
    id = 'comick_pt'
    name = SERVER_NAME
    lang = 'pt'
    lang_code = 'pt'


class Comick_pt_br(Comick):
    id = 'comick_pt_br'
    name = SERVER_NAME
    lang = 'pt_BR'
    lang_code = 'pt-br'


class Comick_ru(Comick):
    id = 'comick_ru'
    name = SERVER_NAME
    lang = 'ru'
    lang_code = 'ru'


class Comick_th(Comick):
    id = 'comick_th'
    name = SERVER_NAME
    lang = 'th'
    lang_code = 'th'


class Comick_uk(Comick):
    id = 'comick_uk'
    name = SERVER_NAME
    lang = 'uk'
    lang_code = 'uk'


class Comick_vi(Comick):
    id = 'comick_vi'
    name = SERVER_NAME
    lang = 'vi'
    lang_code = 'vi'


class Comick_zh_hans(Comick):
    id = 'comick_zh_hans'
    name = SERVER_NAME
    lang = 'zh_Hans'
    lang_code = 'zh'


class Comick_zh_hant(Comick):
    id = 'comick_zh_hant'
    name = SERVER_NAME
    lang = 'zh_Hant'
    lang_code = 'zh-hk'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.my_manga_reader_cms import MyMangaReaderCMS

# Clone of https://mangascan-fr.co


class Bentoscan(MyMangaReaderCMS):
    id = 'bentoscan'
    name = 'Bentoscan'
    lang = 'fr'
    status = 'disabled'

    base_url = 'https://bentoscan.com'
    search_url = base_url + '/search'
    most_populars_url = base_url + '/filterList?page=1&cat=&alpha=&sortBy=views&asc=false&author=&tag=&artist='
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'
    image_url = None  # images all seem to come from a different server (scansmangas.me)
    cover_url = base_url + '/uploads/manga/{0}.jpg'

    details_name_selector = 'h2.widget-title'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Homepage: https://www.kavitareader.com/

import datetime
from functools import cached_property
from gettext import gettext as _
import logging

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import do_login
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

logger = logging.getLogger(__name__)

STATUSES = (
    'ongoing',
    'hiatus',
    'complete',
    'suspended',
    'complete',
)


class Kavita(Server):
    id = 'kavita'
    name = 'Kavita'
    description = _('Self-hosted digital library')
    lang = ''
    has_login = True
    sync = True

    base_url = None  # Customizable via the settings
    logo_url = 'https://www.kavitareader.com/assets/icons/favicon-32x32.png'

    headers = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'User-Agent': 'Komikku Kavita',
    }

    def __init__(self, username=None, password=None, address=None):
        if address:
            self.base_url = address

        if username and password:
            self.do_login(username, password)

    @property
    def api_base_url(self):
        return self.base_url + '/api'

    @property
    def api_chapter_url(self):
        return self.api_base_url + '/Chapter'

    @property
    def api_chapter_read_progress(self):
        return self.api_base_url + '/Reader/progress'

    @property
    def api_chapters_url(self):
        return self.api_base_url + '/Series/series-detail'

    @property
    def api_cover_url(self):
        return self.api_base_url + f'/image/series-cover?seriesId={{{0}}}&apiKey={self.api_key}'

    @property
    def api_image_url(self):
        return self.api_base_url + '/Reader/image'

    @cached_property
    def api_key(self):
        return self.session.cookies.get('apiKey')

    @property
    def api_login_url(self):
        return self.api_base_url + '/Account/login'

    @property
    def api_manga_url(self):
        return self.api_base_url + '/Series/{0}'

    @property
    def api_manga_metadata_url(self):
        return self.api_base_url + '/Series/metadata'

    @property
    def api_search_url(self):
        return self.api_base_url + '/Search/search'

    @property
    def manga_url(self):
        return self.base_url + '/library/{0}/series/{1}'

    @do_login
    def get_manga_data(self, initial_data):
        """
        Returns serie data using API

        Initial data should contain at least serie's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        library_id, series_id = initial_data['slug'].split(':')

        r = self.session_get(
            self.api_manga_url.format(series_id),
            params={
                'seriesId': series_id,
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['name'],
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            last_read=None,
            chapters=[],
            server_id=self.id,
            cover=self.api_cover_url.format(series_id),
        ))

        if latest_read_date := resp_data['latestReadDate']:
            data['last_read'] = datetime.datetime.strptime(latest_read_date[:-1], '%Y-%m-%dT%H:%M:%S.%f')

        r = self.session_get(
            self.api_manga_metadata_url,
            params={
                'seriesId': series_id,
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data['status'] = STATUSES[resp_data['publicationStatus']]

        for key in ('genres', 'tags'):
            for item in resp_data[key]:
                if genre := item.get('name'):
                    data['genres'].append(genre)
                elif genre := item.get('title'):
                    data['genres'].append(genre)

        for key in ('writers',):
            for item in resp_data[key]:
                data['authors'].append(item['name'])

        data['synopsis'] = resp_data['summary']

        # Chapters
        r = self.session_get(
            self.api_chapters_url,
            params={
                'seriesId': series_id,
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        if resp_data['volumes']:
            for volume in resp_data['volumes']:
                for chapter in volume['chapters']:
                    title = f'{volume["name"]} - {chapter["title"]}'
                    if chapter['titleName']:
                        title = f'{title} {chapter["titleName"]}'

                    if modified_date := chapter.get('lastModifiedUtc'):
                        date = modified_date
                    else:
                        date = chapter['createdUtc']

                    if last_reading_progress := chapter.get('lastReadingProgress'):
                        if last_reading_progress.startswith('0001-01-01'):
                            last_read = None
                        else:
                            last_read = datetime.datetime.strptime(last_reading_progress[:-1], '%Y-%m-%dT%H:%M:%S.%f')
                    else:
                        last_read = None

                    chapter_data = dict(
                        slug=f'{volume["id"]}:{chapter["id"]}',  # noqa: E231
                        title=title,
                        num=chapter['number'] if is_number(chapter['number']) else None,
                        date=convert_date_string(date.split('T')[0], format='%Y-%m-%d'),
                        last_read=last_read,
                        last_page_read_index=chapter['pagesRead'] - 1,
                    )
                data['chapters'].append(chapter_data)

        elif resp_data['chapters']:
            # Chapters only, no volumes
            for chapter in resp_data['chapters']:
                if modified_date := chapter.get('lastModifiedUtc'):
                    date = modified_date
                else:
                    date = chapter['createdUtc']

                if last_reading_progress := chapter.get('lastReadingProgress'):
                    if last_reading_progress.startswith('0001-01-01'):
                        last_read = None
                    else:
                        last_read = datetime.datetime.strptime(last_reading_progress[:-1], '%Y-%m-%dT%H:%M:%S.%f')
                else:
                    last_read = None

                chapter_data = dict(
                    slug=f':{chapter["id"]}',  # noqa: E231
                    title=chapter['title'],
                    num=chapter['number'] if is_number(chapter['number']) else None,
                    date=convert_date_string(date.split('T')[0], format='%Y-%m-%d'),
                    last_read=last_read,
                    last_page_read_index=chapter['pagesRead'] - 1,
                )

                data['chapters'].append(chapter_data)

        return data

    @do_login
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns serie chapter data

        Currently, only pages are expected.
        """
        volume_id, chapter_id = chapter_slug.split(':')

        r = self.session_get(
            self.api_chapter_url,
            params={
                'chapterId': chapter_id,
            }
        )
        if r.status_code != 200:
            return None

        data = dict(
            pages=[],
        )
        for index in range(1, r.json()['files'][0]['pages'] + 1):
            data['pages'].append(dict(
                slug=str(index),
                image=None,
            ))

        return data

    @do_login
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        volume_id, chapter_id = chapter_slug.split(':')

        r = self.session_get(
            self.api_image_url,
            params={
                'chapterId': chapter_id,
                'apiKey': self.api_key,
                'page': page['slug'],
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:04d}.{1}'.format(int(page['slug']), mime_type.split('/')[-1]),
        )

    @do_login
    def get_manga_url(self, slug, url):
        """
        Returns serie absolute URL
        """
        library_id, series_id = slug.split(':')

        return self.manga_url.format(library_id, series_id)

    def login(self, username, password):
        try:
            r = self.session.post(
                self.api_login_url,
                json={
                    'username': username,
                    'password': password,
                }
            )
            if r.status_code != 200:
                return False
        except Exception as error:
            logger.warning(error)
            return False

        resp_data = r.json()

        self.token = resp_data['token']
        self.session.cookies.set('apiKey', resp_data['apiKey'], domain=self.base_url)
        self.session.headers.update({'Authorization': f'Bearer {self.token}'})

        self.save_session()

        return True

    @do_login
    def search(self, term):
        r = self.session_get(
            self.api_search_url,
            params={
                'queryString': term,
                'includeChapterAndFiles': 'false',
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['series']:
            if item['format'] in (3, 4):
                # 0: ???
                # 1: Image
                # 2: ???
                # 3: HTML
                # 4: PDF
                continue

            results.append(dict(
                name=item['name'],
                slug=f'{item["libraryId"]}:{item["seriesId"]}',  # noqa: E231
                cover=self.api_cover_url.format(item['seriesId']),
            ))

        return results

    @do_login
    def update_chapter_read_progress(self, data, manga_slug, manga_name, chapter_slug, chapter_url):
        library_id, series_id = manga_slug.split(':')
        volume_id, chapter_id = chapter_slug.split(':')

        r = self.session_post(
            self.api_chapter_read_progress,
            json={
                'volumeId': volume_id,
                'chapterId': chapter_id,
                'pageNum': data['page'],
                'seriesId': series_id,
                'libraryId': library_id,
                # 'bookScrollId': 'string',
                'lastModifiedUtc': datetime.datetime.now(datetime.UTC).strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
            }
        )

        return r.status_code == 200



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class ZeistManga(Server):
    base_url: str

    api_base_url: str = None
    api_search_url: str = None
    api_latest_updates_url: str = None
    api_chapters_url: str = None
    manga_url: str = None
    chapter_url: str = None

    most_popular_selector: str = None
    details_name_selector: str = None
    details_cover_selector: str = None
    details_authors_selector: str = None
    details_type_selector: str = None  # Added in genres
    details_genres_selector: str = None
    details_status_selector: str = None
    details_synopsis_selector: str = None
    chapters_selector: str = None
    chapter_link_selector: str = None
    chapter_title_selector: str = None
    chapter_date_selector: str = None
    pages_selector: str = None

    long_strip_genres = ['Manhua', 'Manhwa']

    def __init__(self):
        if self.api_base_url is None:
            self.api_base_url = self.base_url + '/feeds/posts'
        if self.api_search_url is None:
            self.api_search_url = self.api_base_url + '/default'
        if self.api_latest_updates_url is None:
            self.api_latest_updates_url = self.api_base_url + '/summary/-/Series'
        if self.api_chapters_url is None:
            self.api_chapters_url = self.api_base_url + '/default/-/{0}'
        if self.manga_url is None:
            self.manga_url = self.base_url + '/{0}.html'
        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/{0}.html'

        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = soup.select_one(self.details_name_selector).text.strip()
        data['cover'] = soup.select_one(self.details_cover_selector).get('src')

        if element := soup.select_one(self.details_status_selector):
            status = element.text.strip()
            if status == 'Ongoing':
                data['status'] = 'ongoing'
            elif status == 'Completed':
                data['status'] = 'complete'
            elif status in ('Dropped', 'Cancelled'):
                data['status'] = 'suspended'
            elif status == 'Hiatus':
                data['status'] = 'hiatus'

        for element in soup.select(self.details_authors_selector):
            author = element.text.strip()
            if author in data['authors']:
                continue
            data['authors'].append(author)

        if element := soup.select_one(self.details_type_selector):
            data['genres'].append(element.text.strip())

        for element in soup.select(self.details_genres_selector):
            data['genres'].append(element.text.strip())

        data['synopsis'] = soup.select_one(self.details_synopsis_selector).text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(soup, data)

        return data

    def get_manga_chapters_data(self, soup, data):
        if script_element := soup.select_one('#clwd script'):
            name = script_element.string.split("'")[1]
        else:
            return []

        r = self.session.get(
            self.api_chapters_url.format(name),
            params={
                'alt': 'json-in-script',
                'start-index': 1,
                'max-results': 150,
            },
            headers={
                'Referer': self.get_manga_url(data['slug'], None),
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = json.loads(r.text.replace('gdata.io.handleScriptLoaded(', '')[:-2])

        chapters = []
        for entry in reversed(resp_data['feed']['entry'][1:]):
            link = None
            for link_ in entry['link']:
                if link_['rel'] == 'alternate':
                    link = link_
                    break

            if not link:
                continue

            chapters.append(dict(
                slug=link['href'].replace(f'{self.base_url}/', '').replace('.html', ''),
                title=entry['title']['$t'].strip(),
                date=convert_date_string(entry['updated']['$t'].split('T')[0], '%Y-%m-%d'),
            ))

        return chapters

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for index, img_element in enumerate(soup.select(self.pages_selector), start=1):
            data['pages'].append(dict(
                slug=None,
                image=img_element.get('src'),
                index=index,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0}.{1}'.format(page['index'], mime_type.split('/')[1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return f'{self.base_url}/{slug}'

    def get_latest_updates(self):
        r = self.session_get(
            self.api_latest_updates_url,
            params={
                'alt': 'json-in-script',
                'start-index': 1,
                'max-results': 150,
            },
            headers={
                'Referer': f'{self.base_url}/',
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = json.loads(r.text.replace('gdata.io.handleScriptLoaded(', '')[:-2])

        results = []
        for item in resp_data['feed']['entry']:
            link = None
            for link_ in item['link']:
                if link_['rel'] == 'alternate':
                    link = link_
                    break

            if not link:
                continue

            results.append(dict(
                slug=link['href'].replace(f'{self.base_url}/', '').replace('.html', ''),
                name=link['title'],
                cover=item['media$thumbnail']['url'],
            ))

        return results

    def get_most_populars(self):
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select(self.most_popular_selector):
            results.append(dict(
                slug=a_element.get('href').replace(f'{self.base_url}/', '').replace('.html', ''),
                name=a_element.get('title'),
                cover=a_element.img.get('src'),
            ))

        return results

    def search(self, term):
        r = self.session_get(
            self.api_search_url,
            params={
                'alt': 'json',
                'q': f'label:Series "{term}"',  # noqa E231
                'max-results': 25,
            },
            headers={
                'Referer': f'{self.base_url}/',
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        results = []
        for item in resp_data['feed'].get('entry', []):
            link = None
            for link_ in item['link']:
                if link_['rel'] == 'alternate':
                    link = link_
                    break

            if not link:
                continue

            results.append(dict(
                slug=link['href'].replace(f'{self.base_url}/', '').replace('.html', ''),
                name=link['title'],
                cover=item['media$thumbnail']['url'],
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.manga_stream import MangaStream


class Noromax(MangaStream):
    id = 'noromax'
    name = 'Noromax'
    lang = 'id'

    name_re_sub = r'Bahasa Indonesia'
    series_name = 'manga'

    base_url = 'https://noromax02.my.id'
    logo_url = base_url + '/wp-content/uploads/2026/02/cropped-Untitled-1-150x150.png'

    authors_selector = '.infox .fmed:-soup-contains("Artist") span, .infox .fmed:-soup-contains("Author") span'
    genres_selector = '.infox .mgen a'
    scanlators_selector = '.infox .fmed:-soup-contains("Serialization") span'
    status_selector = '.tsinfo .imptdt:-soup-contains("Status") i'
    synopsis_selector = '[itemprop="description"]'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge


class Phenixscans(Server):
    id = 'phenixscans'
    name = 'Phenix Scans'
    lang = 'fr'

    has_cf = True

    base_url = 'https://phenix-scans.co'
    logo_url = base_url + '/logo.png'
    api_base_url = 'https://api.phenix-scans.co/api'
    api_list_url = api_base_url + '/front/manga'
    api_search_url = api_base_url + '/front/manga/search'
    api_manga_url = api_base_url + '/front/manga/{0}'
    api_chapter_url = api_base_url + '/front/manga/{0}/chapter/{1}'
    manga_url = base_url + '/manga/{0}'
    media_url = 'https://api.phenix-scans.co' + '/{0}'

    filters = [
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': 'Manga', 'name': _('Manga')},
                {'key': 'Manhwa', 'name': _('Manhwa')},
                {'key': 'Manhua', 'name': _('Manhua')},
            ],
        },
    ]
    long_strip_genres = ['Manhua', 'Manhwa']

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data using API

        Initial data should contain at least manga's slug (provided by search)
        """
        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['manga']['title'],
            authors=[],     # Not available
            scanlators=[],  # Not available
            genres=[genre['name'] for genre in resp_data['manga']['genres']],
            status=None,
            synopsis=resp_data['manga'].get('synopsis'),
            chapters=[],
            server_id=self.id,
            cover=self.media_url.format(resp_data['manga']['coverImage']),
        ))

        if resp_data['manga'].get('type'):
            data['genres'].append(resp_data['manga']['type'])

        if resp_data['manga']['status'] == 'Ongoing':
            data['status'] = 'ongoing'
        elif resp_data['manga']['status'] == 'Completed':
            data['status'] = 'complete'
        elif resp_data['manga']['status'] == 'Hiatus':
            data['status'] = 'hiatus'

        # Chapters
        for chapter in reversed(resp_data['chapters']):
            data['chapters'].append({
                'slug': chapter['number'],
                'title': f'Chapitre {chapter["number"]}',
                'num': chapter['number'] if is_number(chapter['number']) else None,
                'date': convert_date_string(chapter['createdAt'].split('T')[0], '%Y-%m-%d'),
            })

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(self.api_chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = dict(
            pages=[],
        )
        for index, image in enumerate(resp_data['chapter']['images']):
            data['pages'].append({
                'slug': None,
                'image': image,
                'index': index + 1,
            })

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.media_url.format(page['image']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_list(self, orderby=None, type=None):
        params = {
            'page': 1,
            'limit': 18,
            'sort': orderby,
        }
        if type:
            params['type'] = type

        r = self.session_get(
            self.api_list_url,
            params=params
        )
        if r.status_code != 200:
            return None

        data = r.json()

        result = []
        for item in data['mangas']:
            result.append(dict(
                name=item['title'],
                slug=item['slug'],
                cover=self.media_url.format(item['coverImage']),
            ))

        return result

    @CompleteChallenge()
    def get_latest_updates(self, type=None):
        return self.get_manga_list(orderby='updatedAt', type=type)

    def get_manga_url(self, slug, url):
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_most_populars(self, type=None):
        return self.get_manga_list(orderby='rating', type=type)

    @CompleteChallenge()
    def search(self, term, type=None):
        # Filtering by type is not available in search endpoint
        r = self.session_get(
            self.api_search_url,
            params={
                'query': term,
            }
        )
        if r.status_code != 200:
            return None

        data = r.json()

        result = []
        for item in data['mangas']:
            result.append(dict(
                name=item['title'],
                slug=item['slug'],
                cover=self.media_url.format(item['coverImage']),
            ))

        return result



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
import json

from bs4 import BeautifulSoup

from komikku.servers.multi.manga_stream import MangaStream
from komikku.utils import get_buffer_mime_type


class Skymangas(MangaStream):
    id = 'skymangas'
    name = 'SkyMangas'
    lang = 'es'
    is_nsfw = True

    base_url = 'https://skymangas.com'
    logo_url = base_url + '/wp-content/uploads/2020/02/cropped-Avatar_SM-1.jpg'

    authors_selector = '.infox .fmed:-soup-contains("Artist") span, .infox .fmed:-soup-contains("Author") span'
    genres_selector = '.infox .mgen a'
    scanlators_selector = '.infox .fmed:-soup-contains("Serialization") span'
    status_selector = '.tsinfo .imptdt:-soup-contains("Status") i'
    synopsis_selector = '[itemprop="description"]'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Pages URLs are encoded in JS
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug=manga_slug, chapter_slug=chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            })
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        # Pages are loaded via javascript
        for script in soup.select('script[src^="data:text/javascript;base64,"]'):
            decoded_pages = base64.b64decode(script.get('src')[28:]).decode()
            if 'ts_reader' not in decoded_pages:
                continue

            pages = json.loads(decoded_pages[14:-2])  # "ts_reader.run(...);"
            for source in pages['sources']:
                # Use first source
                for url in source['images']:
                    data['pages'].append(dict(
                        slug=None,
                        image=url,
                    ))
                break
            break

        return data



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging

from bs4 import BeautifulSoup

from komikku.servers.multi.madtheme import Madtheme
from komikku.servers.utils import convert_date_string
from komikku.webview import CompleteChallenge

logger = logging.getLogger(__name__)


class Kaliscan(Madtheme):
    id = 'kaliscan'
    name = 'KaliScan'
    lang = 'en'

    base_url = 'https://kaliscan.com'
    logo_url = base_url + '/static/sites/icons/favicon-32x32.png'
    api_base_url = base_url + '/service/backend'
    api_chapters_url = api_base_url + '/chaplist/'
    api_chapter_url = api_base_url + '/chapterServer/'

    series_name = 'manga'

    synopsis_selector = 'p.content ~ p'
    images_selector = '.chapter-image'

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        sid = self.get_manga_chapter_id(soup)
        if sid is None:
            logger.warning('Failed to get `%s` chapter server Id', manga_name)
            return None

        r = self.session_get(
            self.api_chapter_url,
            params={
                'server_id': 1,
                'chapter_id': sid,
            },
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for element in soup.select(self.images_selector):
            data['pages'].append(dict(
                slug=None,
                image=element.get(self.image_src_attr),
            ))

        return data

    def get_manga_chapters_data(self, data, soup=None):
        """
        Get list of chapters
        """
        sid = self.get_manga_id(soup)
        if sid is None:
            logger.warning('Failed to get `%s` server Id', data['name'])
            return

        r = self.session_get(
            self.api_chapters_url,
            params={
                'manga_id': sid,
            },
            headers={
                'Referer': self.manga_url.format(data['slug']),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        for element in reversed(soup.select(self.chapters_selector)):
            title_element = element.select_one('.chapter-title')
            date_element = element.select_one('.chapter-update')

            data['chapters'].append({
                'slug': element.get('href').split('/')[-1],
                'title': title_element.text.strip(),
                'date': convert_date_string(date_element.text.strip(), languages=[self.lang]),
            })



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Supported servers:
# Mangareadercc [EN]

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge


class Paprika(Server):
    base_url: str
    search_url: str = None
    latest_updates_url: str = None
    most_populars_url: str = None
    manga_url: str = None
    api_chapters_url: str = None
    chapter_url: str = None

    def __init__(self):
        if self.session is None and not self.has_cf:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

        if self.search_url is None:
            self.search_url = self.base_url + '/search'
        if self.latest_updates_url is None:
            self.latest_updates_url = self.base_url + '/latest-manga'
        if self.most_populars_url is None:
            self.most_populars_url = self.base_url + '/popular-manga'
        if self.manga_url is None:
            self.manga_url = self.base_url + '/manga/{0}'
        if self.api_chapters_url is None:
            self.api_chapters_url = self.base_url + '/ajax-list-chapter?mangaID={0}'
        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/chapter/{0}'

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = soup.select_one('h1').text.strip()
        data['cover'] = soup.select_one('.imgdesc img').get('src')

        if element := soup.select_one('.listinfo li:-soup-contains("Author")'):
            element.b.extract()
            author = element.text[1:].strip()
            data['authors'].append(author)

        if element := soup.select_one('.listinfo li:-soup-contains("Status")'):
            element.b.extract()
            status = element.text[1:].strip()
            if status == 'Ongoing':
                data['status'] = 'ongoing'
            elif status == 'Completed':
                data['status'] = 'complete'

        if element := soup.select_one('.listinfo li:-soup-contains("Genres")'):
            element.b.extract()
            genres = element.text[1:].strip()
            for genre in genres.split(','):
                data['genres'].append(genre.strip())

        if element := soup.select_one('#noidungm'):
            data['synopsis'] = element.text.strip()

        # Chapters
        manga_id = None
        for script_element in soup.find_all('script'):
            script = script_element.string
            if not script or 'mangaID' not in script:
                continue

            for line in script.split('\n'):
                line = line.strip()

                if 'mangaID' in line:
                    manga_id = line.split("'")[-2]
                    break

        if manga_id is None:
            return None

        r = self.session_get(self.api_chapters_url.format(manga_id))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        for element in reversed(soup.select('span.leftoff')):
            title = element.text.strip()
            num = element.a.get('title').split(' ')[-1]  # chapter number theoretically is at end of chapter title

            data['chapters'].append(dict(
                slug=element.a.get('href').split('/')[-1],
                title=title,
                num=num if is_number(num) else None,
                date=None,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for url in soup.select_one('#arraydata').text.split(','):
            data['pages'].append(dict(
                slug=None,
                image=url.strip(),
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_latest_updates(self):
        r = self.session_get(self.latest_updates_url)
        if r.status_code != 200:
            return None

        return self.parse_manga_list(r.text)

    @CompleteChallenge()
    def get_most_populars(self):
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        return self.parse_manga_list(r.text)

    def parse_manga_list(self, html):
        soup = BeautifulSoup(html, 'lxml')

        results = []
        for element in soup.select('.anipost'):
            a_element = element.select_one('.left a')
            results.append(dict(
                slug=a_element.get('href').split('/')[-1],
                name=a_element.text.strip(),
                cover=element.select_one('.thumb img').get('src'),
                last_chapter=element.select_one('span:last-child').text.strip(),
            ))

        return results

    @CompleteChallenge()
    def search(self, term):
        r = self.session_get(
            self.search_url,
            params={
                's': term,
                'post_type': 'manga',
            },
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        return self.parse_manga_list(r.text)
